import copy
import datetime
import json
import logging
import os
import shutil
import tempfile
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import git
import questionary
import rich.prompt
from git.exc import GitCommandError
from typing_extensions import NotRequired, TypedDict  # for py<3.11

import nf_core.utils
from nf_core.components.components_utils import NF_CORE_MODULES_NAME, NF_CORE_MODULES_REMOTE, get_components_to_install
from nf_core.modules.modules_repo import ModulesRepo
from nf_core.pipelines.lint_utils import dump_json_with_prettier

from ..components.components_differ import ComponentsDiffer

log = logging.getLogger(__name__)


class ModulesJsonModuleEntry(TypedDict):
    branch: str
    git_sha: str
    installed_by: List[str]
    patch: NotRequired[str]


class ModulesJsonType(TypedDict):
    name: str
    homePage: str
    repos: Dict[str, Dict[str, Dict[str, Dict[str, ModulesJsonModuleEntry]]]]


class ModulesJson:
    """
    An object for handling a 'modules.json' file in a pipeline
    """

    def __init__(self, pipeline_dir: Union[str, Path]) -> None:
        """
        Initialise the object.

        Args:
            pipeline_dir (str): The pipeline directory
        """
        self.directory = Path(pipeline_dir)
        self.modules_dir = self.directory / "modules"
        self.subworkflows_dir = self.directory / "subworkflows"
        self.modules_json_path = self.directory / "modules.json"
        self.modules_json: Optional[ModulesJsonType] = None
        self.pipeline_modules = None
        self.pipeline_subworkflows = None
        self.pipeline_components: Optional[Dict[str, List[Tuple[str, str]]]] = None

    def __str__(self):
        if self.modules_json is None:
            self.load()
        return json.dumps(self.modules_json, indent=4)

    def __repr__(self):
        return self.__str__()

    def create(self) -> None:
        """
        Creates the modules.json file from the modules and subworkflows installed in the pipeline directory

        Raises:
            UserWarning: If the creation fails
        """
        pipeline_config = nf_core.utils.fetch_wf_config(self.directory)
        pipeline_name = pipeline_config.get("manifest.name", "")
        pipeline_url = pipeline_config.get("manifest.homePage", "")
        new_modules_json = ModulesJsonType(name=pipeline_name, homePage=pipeline_url, repos={})

        if not self.modules_dir.exists():
            if rich.prompt.Confirm.ask(
                "[bold][blue]?[/] Can't find a ./modules directory. Would you like me to create one?", default=True
            ):
                log.info(f"Creating ./modules directory in '{self.directory}'")
                self.modules_dir.mkdir()
            else:
                raise UserWarning("Cannot proceed without a ./modules directory.")

        # Get repositories
        repos, _ = self.get_pipeline_module_repositories("modules", self.modules_dir)
        # Get all module/subworkflow names in the repos
        repo_module_names = self.get_component_names_from_repo(repos, self.modules_dir)
        repo_subworkflow_names = self.get_component_names_from_repo(repos, self.subworkflows_dir)

        # Add module/subworkflow info
        for repo_url, module_names, install_dir in sorted(repo_module_names):
            new_modules_json["repos"][repo_url] = {}
            new_modules_json["repos"][repo_url]["modules"] = {}
            new_modules_json["repos"][repo_url]["modules"][install_dir] = {}
            new_modules_json["repos"][repo_url]["modules"][install_dir] = self.determine_branches_and_shas(
                "modules", install_dir, repo_url, module_names
            )
        for repo_url, subworkflow_names, install_dir in sorted(repo_subworkflow_names):
            if repo_url not in new_modules_json["repos"]:  # Don't overwrite the repo if it was already added by modules
                new_modules_json["repos"][repo_url] = {}
            new_modules_json["repos"][repo_url]["subworkflows"] = {}
            new_modules_json["repos"][repo_url]["subworkflows"][install_dir] = {}
            new_modules_json["repos"][repo_url]["subworkflows"][install_dir] = self.determine_branches_and_shas(
                "subworkflows", install_dir, repo_url, subworkflow_names
            )

        # write the modules.json file and assign it to the object
        self.modules_json = new_modules_json
        self.dump()

    def get_component_names_from_repo(
        self, repos: Dict[str, Dict[str, Dict[str, Dict[str, Dict[str, Union[str, List[str]]]]]]], directory: Path
    ) -> List[Tuple[str, List[str], str]]:
        """
        Get component names from repositories in a pipeline.

        Args:
            repos (list): list of repository urls
            directory (str): modules directory or subworkflows directory

        Returns:
            [(str),[(str),(str)]]: list of tuples with repository url, component names and install directory
        """
        names = []
        for repo_url in repos:
            modules_repo = ModulesRepo(repo_url)
            if modules_repo is None:
                raise UserWarning(f"Could not find module repository for '{repo_url}' in '{directory}'")
            if modules_repo.repo_path is None:
                raise UserWarning(f"Could not find module repository path for '{repo_url}' in '{directory}'")
            components = (
                repo_url,
                [
                    str(Path(component_name).relative_to(directory / modules_repo.repo_path))
                    for component_name, _, file_names in os.walk(directory / modules_repo.repo_path)
                    if "main.nf" in file_names
                ],
                modules_repo.repo_path,
            )
            names.append(components)
        return names

    def get_pipeline_module_repositories(
        self, component_type: str, directory: Path, repos: Optional[Dict] = None
    ) -> Tuple[Dict[str, Dict[str, Dict[str, Dict[str, Dict[str, Union[str, List[str]]]]]]], Dict[Path, Path]]:
        """
        Finds all module repositories in the modules and subworkflows directory.
        Ignores the local modules/subworkflows.

        Args:
            component_type (str): modules or subworkflows
            directory (Path): base directory for the module files
        Returns
            repos ([ (str, str, str) ]),
            renamed_dirs (dict[Path, Path]): List of tuples of repo name, repo
                                             remote URL and path to modules in
                                             repo
        """
        if repos is None:
            repos = {}
        # Check if there are any nf-core modules installed
        if (directory / NF_CORE_MODULES_NAME).exists() and NF_CORE_MODULES_REMOTE not in repos.keys():
            repos[NF_CORE_MODULES_REMOTE] = {}
        # The function might rename some directories, keep track of them
        renamed_dirs = {}
        # Check if there are any untracked repositories

        dirs_not_covered = self.dir_tree_uncovered(directory, [Path(ModulesRepo(url).repo_path) for url in repos])
        if len(dirs_not_covered) > 0:
            log.info(f"Found custom {component_type[:-1]} repositories when creating 'modules.json'")
            # Loop until all directories in the base directory are covered by a remote
            while len(dirs_not_covered) > 0:
                log.info(
                    "The following director{s} in the {t} directory are untracked: '{l}'".format(
                        s="ies" if len(dirs_not_covered) > 0 else "y",
                        t=component_type,
                        l="', '".join(str(dir.relative_to(directory)) for dir in dirs_not_covered),
                    )
                )
                nrepo_remote = questionary.text(
                    "Please provide a URL for for one of the repos contained in the untracked directories.",
                    style=nf_core.utils.nfcore_question_style,
                ).unsafe_ask()
                # Verify that the remote exists
                while True:
                    try:
                        git.Git().ls_remote(nrepo_remote)
                        break
                    except GitCommandError:
                        nrepo_remote = questionary.text(
                            "The provided remote does not seem to exist, please provide a new remote."
                        ).unsafe_ask()

                # Verify that there is a directory corresponding the remote
                nrepo_name = ModulesRepo(nrepo_remote).repo_path
                if nrepo_name is None:
                    raise UserWarning(f"Could not find the repository name for '{nrepo_remote}'")
                if not (directory / nrepo_name).exists():
                    log.info(
                        "The provided remote does not seem to correspond to a local directory. "
                        "The directory structure should be the same as in the remote."
                    )
                    dir_name = questionary.text(
                        "Please provide the correct directory, it will be renamed. If left empty, the remote will be ignored.",
                        style=nf_core.utils.nfcore_question_style,
                    ).unsafe_ask()
                    if dir_name:
                        old_path = directory / dir_name
                        new_path = directory / nrepo_name
                        old_path.rename(new_path)
                        renamed_dirs[old_path] = new_path
                    else:
                        continue

                if nrepo_remote not in repos:
                    repos[nrepo_remote] = {}
                if component_type not in repos[nrepo_remote]:
                    repos[nrepo_remote][component_type] = {}
                repos[nrepo_remote][component_type][nrepo_name] = {}
                dirs_not_covered = self.dir_tree_uncovered(
                    directory, [Path(name) for url in repos for name in repos[url][component_type]]
                )

        return repos, renamed_dirs

    def dir_tree_uncovered(self, components_directory, repos):
        """
        Does a BFS of the modules/subworkflos directory to look for directories that
        are not tracked by a remote. The 'repos' argument contains the
        directories that are currently covered by remote, and it and its
        subdirectories are therefore ignore.

        Args:
            components_directory (Path): Base path of modules or subworkflows in pipeline
            repos ([ Path ]): List of repos that are covered by a remote

        Returns:
            dirs_not_covered ([ Path ]): A list of directories that are currently not covered by any remote.
        """
        # Initialise the FIFO queue. Note that we assume the directory to be correctly
        # configured, i.e. no files etc.
        fifo = [subdir for subdir in components_directory.iterdir() if subdir.stem != "local"]
        depth = 1
        dirs_not_covered = []
        while len(fifo) > 0:
            temp_queue = []
            repos_at_level = {Path(*repo.parts[:depth]): len(repo.parts) for repo in repos}
            for directory in fifo:
                rel_dir = directory.relative_to(components_directory)
                if rel_dir in repos_at_level.keys():
                    # Go the next depth if this directory is not one of the repos
                    if depth < repos_at_level[rel_dir]:
                        temp_queue.extend(directory.iterdir())
                else:
                    # Otherwise add the directory to the ones not covered
                    dirs_not_covered.append(directory)
            fifo = temp_queue
            depth += 1
        return dirs_not_covered

    def determine_branches_and_shas(
        self,
        component_type: str,
        install_dir: Union[str, Path],
        remote_url: str,
        components: List[str],
    ) -> Dict[str, ModulesJsonModuleEntry]:
        """
        Determines what branch and commit sha each module/subworkflow in the pipeline belongs to

        Assumes all modules/subworkflows are installed from the default branch. If it fails to find the
        module/subworkflow in the default branch, it prompts the user with the available branches

        Args:
            install_dir (str): The name of the directory inside modules or subworkflows where components are installed
            remote_url (str): The url to the remote repository
            components ([str]): List of names of installed modules/subworkflows from the repository

        Returns:
            (dict[str, dict[str, str]]): The module.json entries for the modules/subworkflows
                                         from the repository
        """
        default_modules_repo = ModulesRepo(remote_url=remote_url)
        if component_type == "modules":
            repo_path = self.modules_dir / install_dir
        elif component_type == "subworkflows":
            repo_path = self.subworkflows_dir / install_dir
        else:
            raise ValueError(f"Unknown component type '{component_type}'")
        # Get the branches present in the repository, as well as the default branch
        available_branches = ModulesRepo.get_remote_branches(remote_url)
        sb_local = []
        dead_components = []
        repo_entry: Dict[str, ModulesJsonModuleEntry] = {}
        for component in sorted(components):
            modules_repo = default_modules_repo
            component_path = Path(repo_path, component)
            correct_commit_sha = None
            tried_branches = {default_modules_repo.branch}
            found_sha = False
            while True:
                # If the module/subworkflow is patched
                patch_file = component_path / f"{component}.diff"
                if patch_file.is_file():
                    temp_module_dir = self.try_apply_patch_reverse(
                        component_type, component, install_dir, patch_file, component_path
                    )
                    correct_commit_sha = self.find_correct_commit_sha(
                        component_type, component, temp_module_dir, modules_repo
                    )
                else:
                    correct_commit_sha = self.find_correct_commit_sha(
                        component_type, component, component_path, modules_repo
                    )
                    if correct_commit_sha is None:
                        # Check in the old path
                        correct_commit_sha = self.find_correct_commit_sha(
                            component_type, component, repo_path / component_type / component, modules_repo
                        )
                if correct_commit_sha is None:
                    log.info(
                        f"Was unable to find matching {component_type[:-1]} files in the {modules_repo.branch} branch."
                    )
                    choices = [{"name": "No", "value": False}] + [
                        {"name": branch, "value": branch} for branch in (available_branches - tried_branches)
                    ]
                    branch = questionary.select(
                        f"Was the {component_type[:-1]} '{component}' installed from a different branch in the remote?\nSelect 'No' for a local {component_type[:-1]}",
                        choices=choices,
                        style=nf_core.utils.nfcore_question_style,
                    ).unsafe_ask()
                    if not branch:
                        action = questionary.select(
                            f"{component_type[:-1].title()} is untracked '{component}'. Please select what action to take",
                            choices=[
                                {"name": "Move the directory to 'local'", "value": 0},
                                {"name": "Remove the files", "value": 1},
                            ],
                            style=nf_core.utils.nfcore_question_style,
                        ).unsafe_ask()
                        if action == 0:
                            sb_local.append(component)
                        else:
                            dead_components.append(component)
                        break
                    # Create a new modules repo with the selected branch, and retry find the sha
                    modules_repo = ModulesRepo(remote_url=remote_url, branch=branch, no_pull=True, hide_progress=True)
                else:
                    found_sha = True
                    break
            if found_sha and correct_commit_sha is not None:
                repo_entry[component] = {
                    "branch": modules_repo.branch,
                    "git_sha": correct_commit_sha,
                    "installed_by": [component_type],
                }

        # Clean up the modules/subworkflows we were unable to find the sha for
        for component in sb_local:
            log.debug(f"Moving {component_type[:-1]} '{Path(install_dir, component)}' to 'local' directory")
            self.move_component_to_local(component_type, component, str(install_dir))

        for component in dead_components:
            log.debug(f"Removing {component_type[:-1]} {Path(install_dir, component)}'")
            shutil.rmtree(repo_path / component)

        return repo_entry

    def find_correct_commit_sha(
        self,
        component_type: str,
        component_name: Union[str, Path],
        component_path: Union[str, Path],
        modules_repo: ModulesRepo,
    ) -> Optional[str]:
        """
        Returns the SHA for the latest commit where the local files are identical to the remote files
        Args:
            component_type (str): modules or subworkflows
            component_name (str): Name of module/subowrkflow
            component_path (str): Path to module/subworkflow in local repo
            modules_repo (str): Remote repo for module/subworkflow
        Returns:
            commit_sha (str): The latest commit SHA where local files are identical to remote files,
                              or None if no commit is found
        """
        # Find the correct commit SHA for the local module/subworkflow files.
        # We iterate over the commit history for the module/subworkflow until we find
        # a revision that matches the file contents
        commit_shas = (
            commit["git_sha"]
            for commit in modules_repo.get_component_git_log(component_name, component_type, depth=1000)
        )
        for commit_sha in commit_shas:
            if all(
                modules_repo.component_files_identical(
                    component_name, component_path, commit_sha, component_type
                ).values()
            ):
                return commit_sha
        return None

    def move_component_to_local(self, component_type: str, component: str, repo_name: str):
        """
        Move a module/subworkflow to the 'local' directory

        Args:
            component_type (str): The type of component, either 'modules' or 'subworkflows'
            component (str): The name of the module/subworkflow
            repo_name (str): The name of the repository the module resides in
        """
        if component_type == "modules":
            directory = self.modules_dir
        elif component_type == "subworkflows":
            directory = self.subworkflows_dir
        else:
            raise ValueError(f"Unknown component type '{component_type}'")
        current_path = directory / repo_name / component
        local_dir = directory / "local"
        if not local_dir.exists():
            local_dir.mkdir(parents=True)

        to_name = component
        # Check if there is already a subdirectory with the name
        while (local_dir / to_name).exists():
            # Add a time suffix to the path to make it unique
            # (do it again and again if it didn't work out...)
            to_name += f"-{datetime.datetime.now().strftime('%y%m%d%H%M%S')}"
        shutil.move(str(current_path), local_dir / to_name)

    def unsynced_components(self) -> Tuple[List[str], List[str], Dict]:
        """
        Compute the difference between the modules/subworkflows in the directory and the
        modules/subworkflows in the 'modules.json' file. This is done by looking at all
        directories containing a 'main.nf' file

        Returns:
            (untrack_dirs ([ str ]), missing_installation (dict)): Directories that are not tracked
            by the modules.json file, and modules/subworkflows in the modules.json where
            the installation directory is missing
        """
        assert self.modules_json is not None  # mypy
        # Add all modules from modules.json to missing_installation
        missing_installation = copy.deepcopy(self.modules_json["repos"])
        # Obtain the path of all installed modules
        module_dirs = [
            Path(dir_name).relative_to(self.modules_dir)
            for dir_name, _, file_names in os.walk(self.modules_dir)
            if "main.nf" in file_names and not str(Path(dir_name).relative_to(self.modules_dir)).startswith("local")
        ]
        untracked_dirs_modules, missing_installation = self.parse_dirs(module_dirs, missing_installation, "modules")

        # Obtain the path of all installed subworkflows
        subworkflow_dirs = [
            Path(dir_name).relative_to(self.subworkflows_dir)
            for dir_name, _, file_names in os.walk(self.subworkflows_dir)
            if "main.nf" in file_names
            and not str(Path(dir_name).relative_to(self.subworkflows_dir)).startswith("local")
        ]
        untracked_dirs_subworkflows, missing_installation = self.parse_dirs(
            subworkflow_dirs, missing_installation, "subworkflows"
        )

        return untracked_dirs_modules, untracked_dirs_subworkflows, missing_installation

    def parse_dirs(self, dirs: List[Path], missing_installation: Dict, component_type: str) -> Tuple[List[str], Dict]:
        """
        Parse directories and check if they are tracked in the modules.json file

        Args:
            dirs ([ Path ]): List of directories to check
            missing_installation (dict): Dictionary with the modules.json entries
            component_type (str): The type of component, either 'modules' or 'subworkflows'

        Returns:
            (untracked_dirs ([ Path ]), missing_installation (dict)): List of directories that are not tracked
            by the modules.json file, and the updated missing_installation dictionary
        """

        untracked_dirs = []
        for dir_ in dirs:
            # Check if the module/subworkflows directory exists in modules.json
            install_dir = dir_.parts[0]
            component = "/".join(dir_.parts[1:])
            component_in_file = False
            git_url = ""

            for repo in missing_installation:
                if component_type in missing_installation[repo]:
                    if install_dir in missing_installation[repo][component_type]:
                        if component in missing_installation[repo][component_type][install_dir]:
                            component_in_file = True
                            git_url = repo
                            break
            if not component_in_file:
                # If it is not, add it to the list of missing components
                untracked_dirs.append(component)
            else:
                # If it does, remove the component from missing_installation
                module_repo = missing_installation[git_url]
                # Check if the entry has a git sha and branch before removing
                components_dict = module_repo[component_type][install_dir]
                if "git_sha" not in components_dict[component] or "branch" not in components_dict[component]:
                    self.determine_branches_and_shas(component_type, component, git_url, [component])
                # Remove the module/subworkflow from modules/subworkflows without installation
                module_repo[component_type][install_dir].pop(component)
                if len(module_repo[component_type][install_dir]) == 0:
                    # If no modules/subworkflows with missing installation left, remove the install_dir from missing_installation
                    missing_installation[git_url][component_type].pop(install_dir)
                if len(module_repo[component_type]) == 0:
                    # If no modules/subworkflows with missing installation left, remove the component_type from missing_installation
                    missing_installation[git_url].pop(component_type)
                if len(module_repo) == 0:
                    # If no modules/subworkflows with missing installation left, remove the git_url from missing_installation
                    missing_installation.pop(git_url)

        return untracked_dirs, missing_installation

    def has_git_url_and_modules(self) -> bool:
        """
        Check that all repo entries in the modules.json
        has a git url and a modules dict entry
        Returns:
            (bool): True if they are found for all repos, False otherwise
        """
        assert self.modules_json is not None  # mypy
        for repo_url, repo_entry in self.modules_json.get("repos", {}).items():
            if "modules" not in repo_entry:
                if "subworkflows" in repo_entry:
                    continue
                log.warning(f"modules.json entry {repo_entry} does not have a modules entry")
                return False
            elif (
                not isinstance(repo_url, str)
                or repo_url == ""
                or not (
                    repo_url.startswith("http")
                    or repo_url.startswith("ftp")
                    or repo_url.startswith("ssh")
                    or repo_url.startswith("git")
                )
                or not isinstance(repo_entry["modules"], dict)
                or repo_entry["modules"] == {}
            ):
                log.debug(f"modules.json entry {repo_entry} has non-string or empty entries for git_url or modules.")
                return False
        return True

    def reinstall_repo(self, install_dir, remote_url, module_entries):
        """
        Reinstall modules from a repository

        Args:
            install_dir (str): The name of directory where modules are installed
            remote_url (str): The git url of the remote repository
            module_entries ([ dict[str, dict[str, str]] ]): Module entries with
            branch and git sha info

        Returns:
            ([ str ]): List of modules that we failed to install
        """
        branches_and_mods = {}
        failed_to_install = []
        for module, module_entry in module_entries.items():
            if "git_sha" not in module_entry or "branch" not in module_entry:
                failed_to_install.append(module)
            else:
                branch = module_entry["branch"]
                sha = module_entry["git_sha"]
                if branch not in branches_and_mods:
                    branches_and_mods[branch] = []
                branches_and_mods[branch].append((module, sha))

        for branch, modules in branches_and_mods.items():
            try:
                modules_repo = ModulesRepo(remote_url=remote_url, branch=branch)
            except LookupError as e:
                log.error(e)
                failed_to_install.extend(modules)
            for module, sha in modules:
                if not modules_repo.install_component(module, self.modules_dir / install_dir, sha, "modules"):
                    log.warning(
                        f"Could not install module '{Path(self.modules_dir, install_dir, module)}' - removing from modules.json"
                    )
                    failed_to_install.append(module)
        return failed_to_install

    def check_up_to_date(self):
        """
        Checks whether the modules and subworkflows installed in the directory
        are consistent with the entries in the 'modules.json' file and vice versa.

        If a module/subworkflow has an entry in the 'modules.json' file but is missing in the directory,
        we first try to reinstall the module/subworkflow from the remote and if that fails we remove the entry
        in 'modules.json'.

        If a module/subworkflow is installed but the entry in 'modules.json' is missing we iterate through
        the commit log in the remote to try to determine the SHA.

        Check that we have the "installed_by" value in 'modules.json', otherwise add it.
        Assume that the modules/subworkflows were installed by an nf-core command (don't track installed by subworkflows).
        """
        dump_modules_json = False
        try:
            self.load()
            if not self.has_git_url_and_modules():
                raise UserWarning

            assert self.modules_json is not None  # mypy
            # check that all "installed_by" entries are lists and not strings
            # [these strings come from an older dev version, so this check can probably be removed in a future release]
            for _, repo_entry in self.modules_json.get("repos", {}).items():
                for component_type in ["modules", "subworkflows"]:
                    if component_type in repo_entry:
                        for install_dir, install_dir_entry in repo_entry[component_type].items():
                            for _, component in install_dir_entry.items():
                                if "installed_by" in component and isinstance(component["installed_by"], str):
                                    log.debug(f"Updating {component} in modules.json")
                                    dump_modules_json = True
                                    component["installed_by"] = [component["installed_by"]]
        except UserWarning:
            log.info("The 'modules.json' file is not up to date. Recreating the 'modules.json' file.")
            dump_modules_json = True
            self.create()

        # Get unsynced components
        (
            modules_missing_from_modules_json,
            subworkflows_missing_from_modules_json,
            missing_installation,
        ) = self.unsynced_components()

        # If there are any modules/subworkflows left in 'modules.json' after all installed are removed,
        # we try to reinstall them
        if len(missing_installation) > 0:
            if "subworkflows" in [
                c_type for _, repo_content in missing_installation.items() for c_type in repo_content.keys()
            ]:
                self.resolve_missing_installation(missing_installation, "subworkflows")
            if "modules" in [
                c_type for _, repo_content in missing_installation.items() for c_type in repo_content.keys()
            ]:
                self.resolve_missing_installation(missing_installation, "modules")

        # If some modules/subworkflows didn't have an entry in the 'modules.json' file
        # we try to determine the SHA from the commit log of the remote
        if len(modules_missing_from_modules_json) > 0:
            dump_modules_json = True
            self.resolve_missing_from_modules_json(modules_missing_from_modules_json, "modules")
        if len(subworkflows_missing_from_modules_json) > 0:
            dump_modules_json = True
            self.resolve_missing_from_modules_json(subworkflows_missing_from_modules_json, "subworkflows")
        assert self.modules_json is not None  # mypy
        # If the "installed_by" value is not present for modules/subworkflows, add it.
        for repo, repo_content in self.modules_json["repos"].items():
            for component_type, dir_content in repo_content.items():
                for install_dir, installed_components in dir_content.items():
                    for component, component_features in installed_components.items():
                        if "installed_by" not in component_features:
                            dump_modules_json = True
                            self.modules_json["repos"][repo][component_type][install_dir][component]["installed_by"] = [
                                component_type
                            ]

        # Recreate "installed_by" entry
        original_pipeline_components = self.pipeline_components
        self.pipeline_components = None
        subworkflows_dict = self.get_all_components("subworkflows")
        if subworkflows_dict:
            dump_modules_json = True
            for repo, subworkflows in subworkflows_dict.items():
                for org, subworkflow in subworkflows:
                    self.recreate_dependencies(repo, org, subworkflow)
        self.pipeline_components = original_pipeline_components

        if dump_modules_json:
            self.dump(run_prettier=True)
        return True

    def load(self) -> None:
        """
        Loads the modules.json file into the variable 'modules_json'

        Sets the modules_json attribute to the loaded file.

        Raises:
            UserWarning: If the modules.json file is not found
        """
        try:
            with open(self.modules_json_path) as fh:
                try:
                    self.modules_json = json.load(fh)
                except json.JSONDecodeError as e:
                    raise UserWarning(f"Unable to load JSON file '{self.modules_json_path}' due to error {e}")

        except FileNotFoundError:
            raise UserWarning("File 'modules.json' is missing")

    def update(
        self,
        component_type: str,
        modules_repo: ModulesRepo,
        component_name: str,
        component_version: str,
        installed_by: Optional[List[str]],
        installed_by_log: Optional[List[str]] = None,
        write_file: bool = True,
    ) -> bool:
        """
        Updates the 'module.json' file with new module/subworkflow info

        Args:
            component_type (str): modules or subworkflows
            modules_repo (ModulesRepo): A ModulesRepo object configured for the new module/subworkflow
            component_name (str): Name of new module/subworkflow
            component_version (str): git SHA for the new module/subworkflow entry
            installed_by_log (list): previous tracing of installed_by that needs to be added to 'modules.json'
            write_file (bool): whether to write the updated modules.json to a file.

        Returns:
            bool: True if the module/subworkflow was successfully added to the 'modules.json' file
        """
        if installed_by_log is None:
            installed_by_log = []

        if self.modules_json is None:
            self.load()
            assert self.modules_json is not None  # mypy
        repo_name = modules_repo.repo_path
        remote_url = modules_repo.remote_url
        branch = modules_repo.branch

        if remote_url not in self.modules_json["repos"]:
            self.modules_json["repos"][remote_url] = {component_type: {repo_name: {}}}
        if component_type not in self.modules_json["repos"][remote_url]:
            self.modules_json["repos"][remote_url][component_type] = {repo_name: {}}
        repo_component_entry = self.modules_json["repos"][remote_url][component_type][repo_name]
        if component_name not in repo_component_entry:
            repo_component_entry[component_name] = {"branch": "", "git_sha": "", "installed_by": []}
        repo_component_entry[component_name]["git_sha"] = component_version
        repo_component_entry[component_name]["branch"] = branch
        try:
            if installed_by not in repo_component_entry[component_name]["installed_by"] and installed_by is not None:
                repo_component_entry[component_name]["installed_by"] += installed_by
        finally:
            new_installed_by = repo_component_entry[component_name]["installed_by"] + list(installed_by_log)
            repo_component_entry[component_name]["installed_by"] = sorted([*set(new_installed_by)])

        # Sort the 'modules.json' repo entries
        self.modules_json["repos"] = nf_core.utils.sort_dictionary(self.modules_json["repos"])
        if write_file:
            self.dump()
        return True

    def remove_entry(self, component_type, name, repo_url, install_dir, removed_by=None):
        """
        Removes an entry from the 'modules.json' file.

        Args:
            component_type (str): Type of component [modules, subworkflows]
            name (str): Name of the component to be removed
            repo_url (str): URL of the repository containing the component
            install_dir (str): Name of the directory where components are installed
            removed_by (str): Name of the component that wants to remove the component
        Returns:
            (bool): return True if the component was removed, False if it was not found or is still depended on
        """

        if removed_by is None or removed_by == name:
            removed_by = component_type
        if not self.modules_json:
            return False
        if repo_url in self.modules_json.get("repos", {}):
            repo_entry = self.modules_json["repos"][repo_url]
            if name in repo_entry[component_type].get(install_dir, {}):
                if removed_by in repo_entry[component_type][install_dir][name]["installed_by"]:
                    self.modules_json["repos"][repo_url][component_type][install_dir][name]["installed_by"].remove(
                        removed_by
                    )
                    # clean up empty entries
                    if len(repo_entry[component_type][install_dir][name]["installed_by"]) == 0:
                        self.modules_json["repos"][repo_url][component_type][install_dir].pop(name)
                        if len(repo_entry[component_type][install_dir]) == 0:
                            self.modules_json["repos"][repo_url].pop(component_type)
                            if len(repo_entry) == 0:
                                self.modules_json["repos"].pop(repo_url)
                        # write the updated modules.json file
                        self.dump()
                        return True
                    self.dump()
                    return False
            else:
                log.warning(
                    f"{component_type[:-1].title()} '{install_dir}/{name}' is missing from 'modules.json' file."
                )
                return False

        else:
            log.warning(f"{component_type[:-1].title()} '{install_dir}/{name}' is missing from 'modules.json' file.")
            return False

        return False

    def add_patch_entry(self, component_type, component_name, repo_url, install_dir, patch_filename, write_file=True):
        """
        Adds (or replaces) the patch entry for a module
        """
        if self.modules_json is None:
            self.load()
            assert self.modules_json is not None  # mypy

        if repo_url not in self.modules_json["repos"]:
            raise LookupError(f"Repo '{repo_url}' not present in 'modules.json'")
        if component_name not in self.modules_json["repos"][repo_url][component_type][install_dir]:
            raise LookupError(
                f"{component_type[:-1].title()} '{install_dir}/{component_name}' not present in 'modules.json'"
            )
        self.modules_json["repos"][repo_url][component_type][install_dir][component_name]["patch"] = str(patch_filename)
        if write_file:
            self.dump()

    def remove_patch_entry(self, module_name, repo_url, install_dir, write_file=True):
        if self.modules_json is None:
            self.load()
            assert self.modules_json is not None  # mypy

        try:
            del self.modules_json["repos"][repo_url]["modules"][install_dir][module_name]["patch"]
        except KeyError:
            log.warning("No patch entry in 'modules.json' to remove")
        if write_file:
            self.dump()

    def get_patch_fn(self, component_type, component_name, repo_url, install_dir):
        """
        Get the patch filename of a component

        Args:
            component_name (str): The name of the component
            repo_url (str): The URL of the repository containing the component
            install_dir (str): The name of the directory where components are installed

        Returns:
            (str): The patch filename for the component, None if not present
        """
        if self.modules_json is None:
            self.load()
            assert self.modules_json is not None  # mypy
        path = (
            self.modules_json["repos"]
            .get(repo_url, {})
            .get(component_type)
            .get(install_dir)
            .get(component_name, {})
            .get("patch")
        )
        return Path(path) if path is not None else None

    def try_apply_patch_reverse(self, component_type, component, repo_name, patch_relpath, component_dir):
        """
        Try reverse applying a patch file to the modified module or subworkflow files

        Args:
            component_type (str): The type of component [modules, subworkflows]
            component (str): The name of the module or subworkflow
            repo_name (str): The name of the repository where the component resides
            patch_relpath (Path | str): The path to patch file in the pipeline
            component_dir (Path | str): The component directory in the pipeline

        Returns:
            (Path | str): The path of the folder where the component patched files are

        Raises:
            LookupError: If patch was not applied
        """
        component_fullname = str(Path(repo_name, component))
        patch_path = Path(self.directory / patch_relpath)

        try:
            new_files = ComponentsDiffer.try_apply_patch(
                component_type, component, repo_name, patch_path, component_dir, reverse=True
            )
        except LookupError as e:
            raise LookupError(
                f"Failed to apply patch in reverse for {component_type[:-1]} '{component_fullname}' due to: {e}"
            )

        # Write the patched files to a temporary directory
        log.debug("Writing patched files to tmpdir")
        temp_dir = Path(tempfile.mkdtemp())
        temp_component_dir = temp_dir / component
        temp_component_dir.mkdir(parents=True, exist_ok=True)
        for file, new_content in new_files.items():
            fn = temp_component_dir / file
            with open(fn, "w") as fh:
                fh.writelines(new_content)

        return temp_component_dir

    def repo_present(self, repo_name):
        """
        Checks if a repo is present in the modules.json file
        Args:
            repo_name (str): Name of the repository
        Returns:
            (bool): Whether the repo exists in the modules.json
        """
        if self.modules_json is None:
            self.load()
            assert self.modules_json is not None  # mypy

        return repo_name in self.modules_json.get("repos", {})

    def component_present(self, module_name, repo_url, install_dir, component_type):
        """
        Checks if a module is present in the modules.json file
        Args:
            module_name (str): Name of the module
            repo_url (str): URL of the repository
            install_dir (str): Name of the directory where modules are installed
            component_type (str): Type of component [modules, subworkflows]
        Returns:
            (bool): Whether the module is present in the 'modules.json' file
        """
        if self.modules_json is None:
            self.load()
            assert self.modules_json is not None  # mypy
        return module_name in self.modules_json.get("repos", {}).get(repo_url, {}).get(component_type, {}).get(
            install_dir, {}
        )

    def get_modules_json(self) -> ModulesJsonType:
        """
        Returns a copy of the loaded modules.json

        Returns:
            (dict): A copy of the loaded modules.json
        """
        if self.modules_json is None:
            self.load()
            assert self.modules_json is not None  # mypy
        return copy.deepcopy(self.modules_json)

    def get_component_version(self, component_type, component_name, repo_url, install_dir):
        """
        Returns the version of a module or subworkflow

        Args:
            component_name (str): Name of the module/subworkflow
            repo_url (str): URL of the repository
            install_dir (str): Name of the directory where modules/subworkflows are installed

        Returns:
            (str): The git SHA of the module/subworkflow if it exists, None otherwise
        """
        if self.modules_json is None:
            self.load()
            assert self.modules_json is not None  # mypy
        return (
            self.modules_json.get("repos", {})
            .get(repo_url, {})
            .get(component_type, {})
            .get(install_dir, {})
            .get(component_name, {})
            .get("git_sha", None)
        )

    def get_module_version(self, module_name: str, repo_url: str, install_dir: str) -> Optional[str]:
        """
        Returns the version of a module

        Args:
            module_name (str): Name of the module
            repo_url (str): URL of the repository
            install_dir (str): Name of the directory where modules are installed

        Returns:
            (str): The git SHA of the module if it exists, None otherwise
        """
        if self.modules_json is None:
            self.load()
            assert self.modules_json is not None  # mypy
        try:
            sha = self.modules_json["repos"][repo_url]["modules"][install_dir][module_name]["git_sha"]
        except KeyError:
            sha = None
        return sha

    def get_subworkflow_version(self, subworkflow_name, repo_url, install_dir):
        """
        Returns the version of a subworkflow

        Args:
            subworkflow_name (str): Name of the module
            repo_url (str): URL of the repository
            install_dir (str): Name of the directory where subworkflows are installed

        Returns:
            (str): The git SHA of the subworkflow if it exists, None otherwise
        """
        if self.modules_json is None:
            self.load()
            assert self.modules_json is not None  # mypy
        return (
            self.modules_json.get("repos", {})
            .get(repo_url, {})
            .get("subworkflows", {})
            .get(install_dir, {})
            .get(subworkflow_name, {})
            .get("git_sha", None)
        )

    def get_all_components(self, component_type: str) -> Dict[str, List[Tuple[(str, str)]]]:
        """
        Retrieves all pipeline modules/subworkflows that are reported in the modules.json

        Returns:
            (dict[str, [(str, str)]]): Dictionary indexed with the repo urls, with a
                                list of tuples (component_dir, components) as values
        """
        if self.modules_json is None:
            self.load()
            assert self.modules_json is not None  # mypy

        if self.pipeline_components is None:
            self.pipeline_components = {}
            for repo, repo_entry in self.modules_json.get("repos", {}).items():
                if component_type in repo_entry:
                    for directory, components in repo_entry[component_type].items():
                        self.pipeline_components[repo] = [(directory, m) for m in components]

        return self.pipeline_components

    def get_dependent_components(
        self,
        component_type,
        name,
        repo_url,
        install_dir,
        dependent_components,
    ):
        """
        Retrieves all pipeline modules/subworkflows that are reported in the modules.json
        as being installed by the given component

        Args:
            component_type (str): Type of component [modules, subworkflows]
            name (str): Name of the component to find dependencies for
            repo_url (str): URL of the repository containing the components
            install_dir (str): Name of the directory where components are installed

        Returns:
            (dict[str: str,]): Dictionary indexed with the component names, with component_type as value
        """

        if self.modules_json is None:
            self.load()
            assert self.modules_json is not None  # mypy
        component_types = ["modules"] if component_type == "modules" else ["modules", "subworkflows"]
        # Find all components that have an entry of install by of  a given component, recursively call this function for subworkflows
        for type in component_types:
            try:
                components = self.modules_json["repos"][repo_url][type][install_dir].items()
            except KeyError as e:
                # This exception will raise when there are only modules installed
                log.debug(f"Trying to retrieve all {type}. There aren't {type} installed. Failed with error {e}")
                continue
            for component_name, component_entry in components:
                if name in component_entry["installed_by"]:
                    dependent_components[component_name] = type

        return dependent_components

    def get_installed_by_entries(self, component_type, name):
        """
        Retrieves all entries of installed_by for a given component

        Args:
            component_type (str): Type of component [modules, subworkflows]
            name (str): Name of the component to find dependencies for

        Returns:
            (list): The list of installed_by entries

        """
        if self.modules_json is None:
            self.load()
            assert self.modules_json is not None  # mypy
        installed_by_entries = {}
        for _, repo_entry in self.modules_json.get("repos", {}).items():
            if component_type in repo_entry:
                for _, components in repo_entry[component_type].items():
                    if name in components:
                        installed_by_entries = components[name]["installed_by"]
                        break

        return installed_by_entries

    def get_component_branch(self, component_type: str, component: str, repo_url: str, install_dir: str) -> str:
        """
        Gets the branch from which the module/subworkflow was installed

        Returns:
            (str): The branch name
        Raises:
            LookupError: If there is no branch entry in the `modules.json`
        """
        if self.modules_json is None:
            self.load()
            assert self.modules_json is not None  # mypy
        try:
            branch = self.modules_json["repos"][repo_url][component_type][install_dir][component]["branch"]
        except (KeyError, TypeError):
            branch = None
        if branch is None:
            raise LookupError(
                f"Could not find branch information for component '{Path(install_dir, component)}'."
                f"Please remove the 'modules.json' and rerun the command to recreate it"
            )
        return branch

    def dump(self, run_prettier: bool = False) -> None:
        """
        Sort the modules.json, and write it to file
        """
        # Sort the modules.json
        if self.modules_json is None:
            self.load()
        if self.modules_json is not None:
            self.modules_json["repos"] = nf_core.utils.sort_dictionary(self.modules_json["repos"])
            if run_prettier:
                dump_json_with_prettier(self.modules_json_path, self.modules_json)
            else:
                with open(self.modules_json_path, "w") as fh:
                    json.dump(self.modules_json, fh, indent=4)

    def resolve_missing_installation(self, missing_installation: Dict, component_type: str) -> None:
        missing_but_in_mod_json = [
            f"'{component_type}/{install_dir}/{component}'"
            for repo_url, contents in missing_installation.items()
            for install_dir, dir_contents in contents[component_type].items()
            for component in dir_contents
        ]
        log.info(
            f"Reinstalling {component_type} found in 'modules.json' but missing from directory: {', '.join(missing_but_in_mod_json)}"
        )

        remove_from_mod_json = {}
        for repo_url, contents in missing_installation.items():
            for install_dir, component_entries in contents[component_type].items():
                remove_from_mod_json[(repo_url, install_dir)] = self.reinstall_repo(
                    install_dir, repo_url, component_entries
                )

        # If the reinstall fails, we remove those entries in 'modules.json'
        if sum(map(len, remove_from_mod_json.values())) > 0:
            uninstallable_components = [
                f"'{install_dir}/{component}'"
                for (repo_url, install_dir), components in remove_from_mod_json.items()
                for component in components
            ]
            if len(uninstallable_components) == 1:
                log.info(f"Was unable to reinstall {uninstallable_components[0]}. Removing 'modules.json' entry")
            else:
                log.info(
                    f"Was unable to reinstall some {component_type}. Removing 'modules.json' entries: {', '.join(uninstallable_components)}"
                )
            if self.modules_json is None:
                raise UserWarning("No modules.json file found")
            for (repo_url, install_dir), component_entries in remove_from_mod_json.items():
                for component in component_entries:
                    self.modules_json["repos"][repo_url][component_type][install_dir].pop(component)
                if len(self.modules_json["repos"][repo_url][component_type][install_dir]) == 0:
                    self.modules_json["repos"].pop(repo_url)

    def resolve_missing_from_modules_json(self, missing_from_modules_json, component_type):
        format_missing = [f"'{dir}'" for dir in missing_from_modules_json]
        if len(format_missing) == 1:
            log.info(
                f"Recomputing commit SHA for {component_type[:-1]} {format_missing[0]} which was missing from 'modules.json'"
            )
        else:
            log.info(
                f"Recomputing commit SHAs for {component_type} which were missing from 'modules.json': {', '.join(format_missing)}"
            )
        assert self.modules_json is not None  # mypy
        # Get the remotes we are missing
        tracked_repos = {repo_url: (repo_entry) for repo_url, repo_entry in self.modules_json["repos"].items()}
        repos, _ = self.get_pipeline_module_repositories(component_type, self.modules_dir, tracked_repos)

        # Get tuples of components that miss installation and their install directory

        def components_with_repos():
            for directory in missing_from_modules_json:
                for repo_url in repos:
                    modules_repo = ModulesRepo(repo_url)
                    paths_in_directory = []
                    repo_url_path = Path(
                        self.modules_dir,
                        modules_repo.repo_path,
                    )
                    for dir_name, _, _ in os.walk(repo_url_path):
                        if component_type == "modules":
                            if len(Path(directory).parts) > 1:  # The module name is TOOL/SUBTOOL
                                paths_in_directory.append(str(Path(*Path(dir_name).parts[-2:])))
                                pass
                        paths_in_directory.append(Path(dir_name).parts[-1])
                    if directory in paths_in_directory:
                        yield (modules_repo.repo_path, directory)

        # Add all components into a dictionary with install directories
        repos_with_components = {}
        for install_dir, component in components_with_repos():
            if install_dir not in repos_with_components:
                repos_with_components[install_dir] = []
            repos_with_components[install_dir].append(component)

        for install_dir, components in repos_with_components.items():
            remote_url = [
                url
                for url, content in repos.items()
                for comp_type, install_directories in content.items()
                if install_dir in install_directories
            ][0]
            repo_entry = self.determine_branches_and_shas(component_type, install_dir, remote_url, components)
            try:
                self.modules_json["repos"][remote_url][component_type][install_dir].update(repo_entry)
            except KeyError:
                try:
                    self.modules_json["repos"][remote_url][component_type].update({install_dir: repo_entry})
                except KeyError:
                    try:
                        self.modules_json["repos"][remote_url].update(
                            {
                                component_type: {
                                    install_dir: repo_entry,
                                }
                            }
                        )
                    except KeyError:
                        self.modules_json["repos"].update(
                            {
                                remote_url: {
                                    component_type: {
                                        install_dir: repo_entry,
                                    }
                                }
                            }
                        )

    def recreate_dependencies(self, repo, org, subworkflow):
        """
        Try to recreate the installed_by entries for subworkflows.
        Remove self installation entry from dependencies, assuming that the modules.json has been freshly created,
        i.e., no module or subworkflow has been installed by the user in the meantime
        """

        sw_path = Path(self.subworkflows_dir, org, subworkflow)
        dep_mods, dep_subwfs = get_components_to_install(sw_path)
        assert self.modules_json is not None  # mypy
        for dep_mod in dep_mods:
            installed_by = self.modules_json["repos"][repo]["modules"][org][dep_mod]["installed_by"]
            if installed_by == ["modules"]:
                self.modules_json["repos"][repo]["modules"][org][dep_mod]["installed_by"] = []
            if subworkflow not in installed_by:
                self.modules_json["repos"][repo]["modules"][org][dep_mod]["installed_by"].append(subworkflow)

        for dep_subwf in dep_subwfs:
            installed_by = self.modules_json["repos"][repo]["subworkflows"][org][dep_subwf]["installed_by"]
            if installed_by == ["subworkflows"]:
                self.modules_json["repos"][repo]["subworkflows"][org][dep_subwf]["installed_by"] = []
            if subworkflow not in installed_by:
                self.modules_json["repos"][repo]["subworkflows"][org][dep_subwf]["installed_by"].append(subworkflow)
            self.recreate_dependencies(repo, org, dep_subwf)



import filecmp
import logging
import os
import shutil
import tempfile
from pathlib import Path
from typing import Dict, List, Union

import yaml

import nf_core.pipelines.create.create

log = logging.getLogger(__name__)


def files_unchanged(self) -> Dict[str, Union[List[str], bool]]:
    """Checks that certain pipeline files are not modified from template output.

    Iterates through the pipeline's directory content and compares specified files
    against output from the template using the pipeline's metadata. File content
    should not be modified / missing.

    Files that must be unchanged::

        .gitattributes
        .prettierrc.yml
        .github/.dockstore.yml
        .github/CONTRIBUTING.md
        .github/ISSUE_TEMPLATE/bug_report.yml
        .github/ISSUE_TEMPLATE/config.yml
        .github/ISSUE_TEMPLATE/feature_request.yml
        .github/PULL_REQUEST_TEMPLATE.md
        .github/workflows/branch.yml
        .github/workflows/linting_comment.yml
        .github/workflows/linting.yml
        assets/email_template.html
        assets/email_template.txt
        assets/nf-core-PIPELINE_logo_light.png
        assets/sendmail_template.txt
        CODE_OF_CONDUCT.md
        docs/images/nf-core-PIPELINE_logo_light.png
        docs/images/nf-core-PIPELINE_logo_dark.png
        docs/README.md'
        ['LICENSE', 'LICENSE.md', 'LICENCE', 'LICENCE.md'], # NB: British / American spelling

    Files that can have additional content but must include the template contents::

        .gitignore
        .prettierignore


    .. tip:: You can configure the ``nf-core pipelines lint`` tests to ignore any of these checks by setting
             the ``files_unchanged`` key as follows in your ``.nf-core.yml`` config file. For example:

             .. code-block:: yaml

                lint:
                    files_unchanged:
                        - .github/workflows/branch.yml

    """

    passed: List[str] = []
    failed: List[str] = []
    warned: List[str] = []
    ignored: List[str] = []
    fixed: List[str] = []
    could_fix: bool = False

    # Check that we have the minimum required config
    required_pipeline_config = {"manifest.name", "manifest.description", "manifest.author"}
    missing_pipeline_config = required_pipeline_config.difference(self.nf_config)
    if missing_pipeline_config:
        return {"ignored": [f"Required pipeline config not found - {missing_pipeline_config}"]}
    try:
        prefix, short_name = self.nf_config["manifest.name"].strip("\"'").split("/")
    except ValueError:
        log.warning(
            "Expected manifest.name to be in the format '<repo>/<pipeline>'. Will assume it is <pipeline> and default to repo 'nf-core'"
        )
        short_name = self.nf_config["manifest.name"].strip("\"'")
        prefix = "nf-core"

    # NB: Should all be files, not directories
    # List of lists. Passes if any of the files in the sublist are found.
    files_exact = [
        [Path(".gitattributes")],
        [Path(".prettierrc.yml")],
        [Path("CODE_OF_CONDUCT.md")],
        [Path("LICENSE"), Path("LICENSE.md"), Path("LICENCE"), Path("LICENCE.md")],  # NB: British / American spelling
        [Path(".github", ".dockstore.yml")],
        [Path(".github", "CONTRIBUTING.md")],
        [Path(".github", "ISSUE_TEMPLATE", "bug_report.yml")],
        [Path(".github", "ISSUE_TEMPLATE", "config.yml")],
        [Path(".github", "ISSUE_TEMPLATE", "feature_request.yml")],
        [Path(".github", "PULL_REQUEST_TEMPLATE.md")],
        [Path(".github", "workflows", "branch.yml")],
        [Path(".github", "workflows", "linting_comment.yml")],
        [Path(".github", "workflows", "linting.yml")],
        [Path("assets", "email_template.html")],
        [Path("assets", "email_template.txt")],
        [Path("assets", "sendmail_template.txt")],
        [Path("assets", f"nf-core-{short_name}_logo_light.png")],
        [Path("docs", "images", f"nf-core-{short_name}_logo_light.png")],
        [Path("docs", "images", f"nf-core-{short_name}_logo_dark.png")],
        [Path("docs", "README.md")],
    ]
    files_partial = [
        [Path(".gitignore"), Path(".prettierignore")],
    ]

    # Only show error messages from pipeline creation
    logging.getLogger("nf_core.pipelines.create").setLevel(logging.ERROR)

    # Generate a new pipeline with nf-core create that we can compare to
    tmp_dir = Path(tempfile.TemporaryDirectory().name)
    tmp_dir.mkdir(parents=True)

    # Create a template.yaml file for the pipeline creation
    template_yaml = {
        "name": short_name,
        "description": self.nf_config["manifest.description"].strip("\"'"),
        "author": self.nf_config["manifest.author"].strip("\"'"),
        "org": prefix,
    }

    template_yaml_path = Path(tmp_dir, "template.yaml")

    with open(template_yaml_path, "w") as fh:
        yaml.dump(template_yaml, fh, default_flow_style=False)

    test_pipeline_dir = Path(tmp_dir, f"{prefix}-{short_name}")
    create_obj = nf_core.pipelines.create.create.PipelineCreate(
        None, None, None, no_git=True, outdir=test_pipeline_dir, template_config=template_yaml_path
    )
    create_obj.init_pipeline()

    # Helper functions for file paths
    def _pf(file_path: Union[str, Path]) -> Path:
        """Helper function - get file path for pipeline file"""
        return Path(self.wf_path, file_path)

    def _tf(file_path: Union[str, Path]) -> Path:
        """Helper function - get file path for template file"""
        return Path(test_pipeline_dir, file_path)

    ignore_files = self.lint_config.get("files_unchanged", []) if self.lint_config is not None else []

    # Files that must be completely unchanged from template
    for files in files_exact:
        # Ignore if file specified in linting config
        if any([str(f) in ignore_files for f in files]):
            ignored.append(f"File ignored due to lint config: {self._wrap_quotes(files)}")

        # Ignore if we can't find the file
        elif not any([_pf(f).is_file() for f in files]):
            ignored.append(f"File does not exist: {self._wrap_quotes(files)}")

        # Check that the file has an identical match
        else:
            for f in files:
                try:
                    if filecmp.cmp(_pf(f), _tf(f), shallow=True):
                        passed.append(f"`{f}` matches the template")
                    else:
                        if f.name.endswith(".png") and int(os.stat(_pf(f)).st_size / 500) == int(
                            os.stat(_tf(f)).st_size / 500
                        ):
                            # almost the same file, good enough for the logo
                            log.debug(f"Files are almost the same. Will pass: {f}")
                            passed.append(f"`{f}` matches the template")
                        elif "files_unchanged" in self.fix:
                            # Try to fix the problem by overwriting the pipeline file
                            shutil.copy(_tf(f), _pf(f))
                            passed.append(f"`{f}` matches the template")
                            fixed.append(f"`{f}` overwritten with template file")
                        elif f.name in ["LICENSE", "LICENSE.md", "LICENCE", "LICENCE.md"]:
                            # Report LICENSE as a warning since we are not using the manifest.author names
                            # TODO: Lint the content of the LICENSE file except the line containing author names
                            # to allow for people to opt-in listing author/maintainer names instead of using the "nf-core community"
                            warned.append(f"`{f}` does not match the template")
                            could_fix = True
                        else:
                            failed.append(f"`{f}` does not match the template")
                            could_fix = True
                except FileNotFoundError:
                    pass

    # Files that can be added to, but that must contain the template contents
    for files in files_partial:
        # Ignore if file specified in linting config
        if any([str(f) in ignore_files for f in files]):
            ignored.append(f"File ignored due to lint config: {self._wrap_quotes(files)}")

        # Ignore if we can't find the file
        elif not any([_pf(f).is_file() for f in files]):
            ignored.append(f"File does not exist: {self._wrap_quotes(files)}")

        # Check that the file contains the template file contents
        else:
            for f in files:
                try:
                    with open(_pf(f)) as fh:
                        pipeline_file = fh.read()
                    with open(_tf(f)) as fh:
                        template_file = fh.read()
                    if template_file in pipeline_file:
                        passed.append(f"`{f}` matches the template")
                    else:
                        if "files_unchanged" in self.fix:
                            # Try to fix the problem by overwriting the pipeline file
                            with open(_tf(f)) as fh:
                                template_file = fh.read()
                            with open(_pf(f), "w") as fh:
                                fh.write(template_file)
                            passed.append(f"`{f}` matches the template")
                            fixed.append(f"`{f}` overwritten with template file")
                        else:
                            failed.append(f"`{f}` does not match the template")
                            could_fix = True
                except FileNotFoundError:
                    pass

    # cleaning up temporary dir
    shutil.rmtree(tmp_dir)

    return {
        "passed": passed,
        "failed": failed,
        "warned": warned,
        "ignored": ignored,
        "fixed": fixed,
        "could_fix": could_fix,
    }



from textwrap import dedent

from textual import on
from textual.app import ComposeResult
from textual.containers import Center, Horizontal, ScrollableContainer
from textual.screen import Screen
from textual.widgets import Button, Footer, Header, Markdown, Static, Switch

from nf_core.pipelines.create.utils import PipelineFeature


class CustomPipeline(Screen):
    """Select if the pipeline will use genomic data."""

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        yield Markdown(
            dedent(
                """
                # Template features
                """
            )
        )
        yield Horizontal(
            Switch(id="toggle_all", value=True),
            Static("Toggle all features", classes="feature_title"),
            classes="custom_grid",
        )
        yield ScrollableContainer(id="features")

        yield Center(
            Button("Back", id="back", variant="default"),
            Button("Continue", id="continue", variant="success"),
            classes="cta",
        )

    def on_mount(self) -> None:
        for name, feature in self.parent.template_features_yml.items():
            if feature["custom_pipelines"]:
                self.query_one("#features").mount(
                    PipelineFeature(feature["help_text"], feature["short_description"], feature["description"], name)
                )
        self.query_one("#toggle_all", Switch).value = True

    @on(Button.Pressed, "#continue")
    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Save fields to the config."""
        skip = []
        for feature_input in self.query("PipelineFeature"):
            this_switch = feature_input.query_one(Switch)
            if not this_switch.value:
                skip.append(this_switch.id)
        self.parent.TEMPLATE_CONFIG.__dict__.update({"skip_features": skip, "is_nfcore": False})

    @on(Switch.Changed, "#toggle_all")
    def on_toggle_all(self, event: Switch.Changed) -> None:
        """Handle toggling all switches."""
        new_state = event.value
        for feature in self.query("PipelineFeature"):
            feature.query_one(Switch).value = new_state



import json
from pathlib import Path
from typing import Union

import yaml
from git.repo import Repo

import nf_core.modules.lint
import nf_core.modules.patch
from nf_core.modules.lint.main_nf import check_container_link_line, check_process_labels
from nf_core.utils import set_wd

from ..test_modules import TestModules
from ..utils import GITLAB_NFTEST_BRANCH, GITLAB_URL
from .test_patch import BISMARK_ALIGN, CORRECT_SHA, PATCH_BRANCH, REPO_NAME, modify_main_nf

PROCESS_LABEL_GOOD = (
    """
    label 'process_high'
    cpus 12
    """,
    1,
    0,
    0,
)
PROCESS_LABEL_NON_ALPHANUMERIC = (
    """
    label 'a:label:with:colons'
    cpus 12
    """,
    0,
    2,
    0,
)
PROCESS_LABEL_GOOD_CONFLICTING = (
    """
    label 'process_high'
    label 'process_low'
    cpus 12
    """,
    0,
    1,
    0,
)
PROCESS_LABEL_GOOD_DUPLICATES = (
    """
    label 'process_high'
    label 'process_high'
    cpus 12
    """,
    0,
    2,
    0,
)
PROCESS_LABEL_GOOD_AND_NONSTANDARD = (
    """
    label 'process_high'
    label 'process_extra_label'
    cpus 12
    """,
    1,
    1,
    0,
)
PROCESS_LABEL_NONSTANDARD = (
    """
    label 'process_extra_label'
    cpus 12
    """,
    0,
    2,
    0,
)
PROCESS_LABEL_NONSTANDARD_DUPLICATES = (
    """
    label process_extra_label
    label process_extra_label
    cpus 12
    """,
    0,
    3,
    0,
)
PROCESS_LABEL_NONE_FOUND = (
    """
    cpus 12
    """,
    0,
    1,
    0,
)

PROCESS_LABEL_TEST_CASES = [
    PROCESS_LABEL_GOOD,
    PROCESS_LABEL_NON_ALPHANUMERIC,
    PROCESS_LABEL_GOOD_CONFLICTING,
    PROCESS_LABEL_GOOD_DUPLICATES,
    PROCESS_LABEL_GOOD_AND_NONSTANDARD,
    PROCESS_LABEL_NONSTANDARD,
    PROCESS_LABEL_NONSTANDARD_DUPLICATES,
    PROCESS_LABEL_NONE_FOUND,
]


# Test cases for linting the container definitions

CONTAINER_SINGLE_GOOD = (
    "Single-line container definition should pass",
    """
    container "quay.io/nf-core/gatk:4.4.0.0" //Biocontainers is missing a package
    """,
    2,  # passed
    0,  # warned
    0,  # failed
)

CONTAINER_TWO_LINKS_GOOD = (
    "Multi-line container definition should pass",
    """
    container "${ workflow.containerEngine == 'singularity' && !task.ext.singularity_pull_docker_container ?
        'https://depot.galaxyproject.org/singularity/gatk4:4.4.0.0--py36hdfd78af_0':
        'biocontainers/gatk4:4.4.0.0--py36hdfd78af_0' }"
    """,
    6,
    0,
    0,
)

CONTAINER_WITH_SPACE_BAD = (
    "Space in container URL should fail",
    """
    container "${ workflow.containerEngine == 'singularity' && !task.ext.singularity_pull_docker_container ?
        'https://depot.galaxyproject.org/singularity/gatk4:4.4.0.0--py36hdfd78af_0 ':
        'biocontainers/gatk4:4.4.0.0--py36hdfd78af_0' }"
    """,
    5,
    0,
    1,
)

CONTAINER_MULTIPLE_DBLQUOTES_BAD = (
    "Incorrect quoting of container string should fail",
    """
    container "${ workflow.containerEngine == 'singularity' && !task.ext.singularity_pull_docker_container ?
        'https://depot.galaxyproject.org/singularity/gatk4:4.4.0.0--py36hdfd78af_0 ':
        "biocontainers/gatk4:4.4.0.0--py36hdfd78af_0" }"
    """,
    4,
    0,
    1,
)

CONTAINER_TEST_CASES = [
    CONTAINER_SINGLE_GOOD,
    CONTAINER_TWO_LINKS_GOOD,
    CONTAINER_WITH_SPACE_BAD,
    CONTAINER_MULTIPLE_DBLQUOTES_BAD,
]


class TestModulesCreate(TestModules):
    def _setup_patch(self, pipeline_dir: Union[str, Path], modify_module: bool):
        install_obj = nf_core.modules.install.ModuleInstall(
            pipeline_dir,
            prompt=False,
            force=False,
            remote_url=GITLAB_URL,
            branch=PATCH_BRANCH,
            sha=CORRECT_SHA,
        )

        # Install the module
        install_obj.install(BISMARK_ALIGN)

        if modify_module:
            # Modify the module
            module_path = Path(pipeline_dir, "modules", REPO_NAME, BISMARK_ALIGN)
            modify_main_nf(module_path / "main.nf")

    def test_modules_lint_trimgalore(self):
        """Test linting the TrimGalore! module"""
        self.mods_install.install("trimgalore")
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.pipeline_dir)
        module_lint.lint(print_results=False, module="trimgalore")
        assert len(module_lint.failed) == 0, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) > 0
        assert len(module_lint.warned) >= 0

    def test_modules_lint_tabix_tabix(self):
        """Test linting the tabix/tabix module"""
        self.mods_install.install("tabix/tabix")
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.pipeline_dir)
        module_lint.lint(print_results=False, module="tabix/tabix")
        assert len(module_lint.failed) == 0, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) > 0
        assert len(module_lint.warned) >= 0

    def test_modules_lint_empty(self):
        """Test linting a pipeline with no modules installed"""
        self.mods_remove.remove("fastqc", force=True)
        self.mods_remove.remove("multiqc", force=True)
        nf_core.modules.lint.ModuleLint(directory=self.pipeline_dir)
        assert "No modules from https://github.com/nf-core/modules.git installed in pipeline" in self.caplog.text

    def test_modules_lint_new_modules(self):
        """lint a new module"""
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.nfcore_modules)
        module_lint.lint(print_results=False, all_modules=True)
        assert len(module_lint.failed) == 0, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) > 0
        assert len(module_lint.warned) >= 0

    def test_modules_lint_no_gitlab(self):
        """Test linting a pipeline with no modules installed"""
        self.mods_remove.remove("fastqc", force=True)
        self.mods_remove.remove("multiqc", force=True)
        nf_core.modules.lint.ModuleLint(directory=self.pipeline_dir, remote_url=GITLAB_URL)
        assert f"No modules from {GITLAB_URL} installed in pipeline" in self.caplog.text

    def test_modules_lint_gitlab_modules(self):
        """Lint modules from a different remote"""
        self.mods_install_gitlab.install("fastqc")
        self.mods_install_gitlab.install("multiqc")
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.pipeline_dir, remote_url=GITLAB_URL)
        module_lint.lint(print_results=False, all_modules=True)
        assert len(module_lint.failed) == 2
        assert len(module_lint.passed) > 0
        assert len(module_lint.warned) >= 0

    def test_modules_lint_multiple_remotes(self):
        """Lint modules from a different remote"""
        self.mods_install_gitlab.install("multiqc")
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.pipeline_dir, remote_url=GITLAB_URL)
        module_lint.lint(print_results=False, all_modules=True)
        assert len(module_lint.failed) == 1
        assert len(module_lint.passed) > 0
        assert len(module_lint.warned) >= 0

    def test_modules_lint_registry(self):
        """Test linting the samtools module and alternative registry"""
        assert self.mods_install.install("samtools/sort")
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.pipeline_dir, registry="public.ecr.aws")
        module_lint.lint(print_results=False, module="samtools/sort")
        assert len(module_lint.failed) == 0, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) > 0
        assert len(module_lint.warned) >= 0
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.pipeline_dir)
        module_lint.lint(print_results=False, module="samtools/sort")
        assert len(module_lint.failed) == 0, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) > 0
        assert len(module_lint.warned) >= 0

    def test_modules_lint_patched_modules(self):
        """
        Test creating a patch file and applying it to a new version of the the files
        """
        self._setup_patch(str(self.pipeline_dir), True)

        # Create a patch file
        patch_obj = nf_core.modules.patch.ModulePatch(self.pipeline_dir, GITLAB_URL, PATCH_BRANCH)
        patch_obj.patch(BISMARK_ALIGN)

        # change temporarily working directory to the pipeline directory
        # to avoid error from try_apply_patch() during linting
        with set_wd(self.pipeline_dir):
            module_lint = nf_core.modules.lint.ModuleLint(
                directory=self.pipeline_dir,
                remote_url=GITLAB_URL,
                branch=PATCH_BRANCH,
                hide_progress=True,
            )
            module_lint.lint(
                print_results=False,
                all_modules=True,
            )

        assert len(module_lint.failed) == 1, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) > 0
        assert len(module_lint.warned) >= 0

    def test_modules_lint_check_process_labels(self):
        for test_case in PROCESS_LABEL_TEST_CASES:
            process, passed, warned, failed = test_case
            mocked_ModuleLint = MockModuleLint()
            check_process_labels(mocked_ModuleLint, process.splitlines())
            assert len(mocked_ModuleLint.passed) == passed
            assert len(mocked_ModuleLint.warned) == warned
            assert len(mocked_ModuleLint.failed) == failed

    def test_modules_lint_check_url(self):
        for test_case in CONTAINER_TEST_CASES:
            test, process, passed, warned, failed = test_case
            mocked_ModuleLint = MockModuleLint()
            for line in process.splitlines():
                if line.strip():
                    check_container_link_line(mocked_ModuleLint, line, registry="quay.io")

            assert (
                len(mocked_ModuleLint.passed) == passed
            ), f"{test}: Expected {passed} PASS, got {len(mocked_ModuleLint.passed)}."
            assert (
                len(mocked_ModuleLint.warned) == warned
            ), f"{test}: Expected {warned} WARN, got {len(mocked_ModuleLint.warned)}."
            assert (
                len(mocked_ModuleLint.failed) == failed
            ), f"{test}: Expected {failed} FAIL, got {len(mocked_ModuleLint.failed)}."

    def test_modules_lint_update_meta_yml(self):
        """update the meta.yml of a module"""
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.nfcore_modules, fix=True)
        module_lint.lint(print_results=False, module="bpipe/test")
        assert len(module_lint.failed) == 0, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) > 0
        assert len(module_lint.warned) >= 0

    def test_modules_lint_snapshot_file(self):
        """Test linting a module with a snapshot file"""
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.nfcore_modules)
        module_lint.lint(print_results=False, module="bpipe/test")
        assert len(module_lint.failed) == 0, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) > 0
        assert len(module_lint.warned) >= 0

    def test_modules_lint_snapshot_file_missing_fail(self):
        """Test linting a module with a snapshot file missing, which should fail"""
        Path(
            self.nfcore_modules,
            "modules",
            "nf-core",
            "bpipe",
            "test",
            "tests",
            "main.nf.test.snap",
        ).unlink()
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.nfcore_modules)
        module_lint.lint(print_results=False, module="bpipe/test")
        Path(
            self.nfcore_modules,
            "modules",
            "nf-core",
            "bpipe",
            "test",
            "tests",
            "main.nf.test.snap",
        ).touch()
        assert len(module_lint.failed) == 1, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) > 0
        assert len(module_lint.warned) >= 0
        assert module_lint.failed[0].lint_test == "test_snapshot_exists"

    def test_modules_lint_snapshot_file_not_needed(self):
        """Test linting a module which doesn't need a snapshot file by removing the snapshot keyword in the main.nf.test file"""
        with open(
            Path(
                self.nfcore_modules,
                "modules",
                "nf-core",
                "bpipe",
                "test",
                "tests",
                "main.nf.test",
            )
        ) as fh:
            content = fh.read()
            new_content = content.replace("snapshot(", "snap (")
        with open(
            Path(
                self.nfcore_modules,
                "modules",
                "nf-core",
                "bpipe",
                "test",
                "tests",
                "main.nf.test",
            ),
            "w",
        ) as fh:
            fh.write(new_content)
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.nfcore_modules)
        module_lint.lint(print_results=False, module="bpipe/test")
        assert len(module_lint.failed) == 0, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) > 0
        assert len(module_lint.warned) >= 0

    def test_modules_environment_yml_file_doesnt_exists(self):
        """Test linting a module with an environment.yml file"""
        Path(self.nfcore_modules, "modules", "nf-core", "bpipe", "test", "environment.yml").rename(
            Path(
                self.nfcore_modules,
                "modules",
                "nf-core",
                "bpipe",
                "test",
                "environment.yml.bak",
            )
        )
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.nfcore_modules)
        module_lint.lint(print_results=False, module="bpipe/test")
        Path(
            self.nfcore_modules,
            "modules",
            "nf-core",
            "bpipe",
            "test",
            "environment.yml.bak",
        ).rename(
            Path(
                self.nfcore_modules,
                "modules",
                "nf-core",
                "bpipe",
                "test",
                "environment.yml",
            )
        )
        assert len(module_lint.failed) == 1, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) > 0
        assert len(module_lint.warned) >= 0
        assert module_lint.failed[0].lint_test == "environment_yml_exists"

    def test_modules_environment_yml_file_sorted_correctly(self):
        """Test linting a module with a correctly sorted environment.yml file"""
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.nfcore_modules)
        module_lint.lint(print_results=False, module="bpipe/test")
        assert len(module_lint.failed) == 0, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) > 0
        assert len(module_lint.warned) >= 0

    def test_modules_environment_yml_file_sorted_incorrectly(self):
        """Test linting a module with an incorrectly sorted environment.yml file"""
        with open(
            Path(
                self.nfcore_modules,
                "modules",
                "nf-core",
                "bpipe",
                "test",
                "environment.yml",
            )
        ) as fh:
            yaml_content = yaml.safe_load(fh)
        # Add a new dependency to the environment.yml file and reverse the order
        yaml_content["dependencies"].append("z=0.0.0")
        yaml_content["dependencies"].reverse()
        yaml_content = yaml.dump(yaml_content)
        with open(
            Path(
                self.nfcore_modules,
                "modules",
                "nf-core",
                "bpipe",
                "test",
                "environment.yml",
            ),
            "w",
        ) as fh:
            fh.write(yaml_content)
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.nfcore_modules)
        module_lint.lint(print_results=False, module="bpipe/test")
        # we fix the sorting on the fly, so this should pass
        assert len(module_lint.failed) == 0, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) > 0
        assert len(module_lint.warned) >= 0

    def test_modules_environment_yml_file_not_array(self):
        """Test linting a module with an incorrectly formatted environment.yml file"""
        with open(
            Path(
                self.nfcore_modules,
                "modules",
                "nf-core",
                "bpipe",
                "test",
                "environment.yml",
            )
        ) as fh:
            yaml_content = yaml.safe_load(fh)
        yaml_content["dependencies"] = "z"
        with open(
            Path(
                self.nfcore_modules,
                "modules",
                "nf-core",
                "bpipe",
                "test",
                "environment.yml",
            ),
            "w",
        ) as fh:
            fh.write(yaml.dump(yaml_content))
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.nfcore_modules)
        module_lint.lint(print_results=False, module="bpipe/test")
        assert len(module_lint.failed) == 1, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) > 0
        assert len(module_lint.warned) >= 0
        assert module_lint.failed[0].lint_test == "environment_yml_valid"

    def test_modules_meta_yml_incorrect_licence_field(self):
        """Test linting a module with an incorrect Licence field in meta.yml"""
        with open(Path(self.nfcore_modules, "modules", "nf-core", "bpipe", "test", "meta.yml")) as fh:
            meta_yml = yaml.safe_load(fh)
        meta_yml["tools"][0]["bpipe"]["licence"] = "[MIT]"
        with open(
            Path(self.nfcore_modules, "modules", "nf-core", "bpipe", "test", "meta.yml"),
            "w",
        ) as fh:
            fh.write(yaml.dump(meta_yml))
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.nfcore_modules)
        module_lint.lint(print_results=False, module="bpipe/test")

        # reset changes
        meta_yml["tools"][0]["bpipe"]["licence"] = ["MIT"]
        with open(
            Path(self.nfcore_modules, "modules", "nf-core", "bpipe", "test", "meta.yml"),
            "w",
        ) as fh:
            fh.write(yaml.dump(meta_yml))

        assert len(module_lint.failed) == 1, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) >= 0
        assert len(module_lint.warned) >= 0
        assert module_lint.failed[0].lint_test == "meta_yml_valid"

    def test_modules_meta_yml_output_mismatch(self):
        """Test linting a module with an extra entry in output fields in meta.yml compared to module.output"""
        with open(Path(self.nfcore_modules, "modules", "nf-core", "bpipe", "test", "main.nf")) as fh:
            main_nf = fh.read()
        main_nf_new = main_nf.replace("emit: bam", "emit: bai")
        with open(Path(self.nfcore_modules, "modules", "nf-core", "bpipe", "test", "main.nf"), "w") as fh:
            fh.write(main_nf_new)
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.nfcore_modules)
        module_lint.lint(print_results=False, module="bpipe/test")
        with open(Path(self.nfcore_modules, "modules", "nf-core", "bpipe", "test", "main.nf"), "w") as fh:
            fh.write(main_nf)
        assert len(module_lint.failed) == 1, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) >= 0
        assert "Module `meta.yml` does not match `main.nf`" in module_lint.failed[0].message

    def test_modules_meta_yml_incorrect_name(self):
        """Test linting a module with an incorrect name in meta.yml"""
        with open(Path(self.nfcore_modules, "modules", "nf-core", "bpipe", "test", "meta.yml")) as fh:
            meta_yml = yaml.safe_load(fh)
        meta_yml["name"] = "bpipe/test"
        with open(
            Path(self.nfcore_modules, "modules", "nf-core", "bpipe", "test", "meta.yml"),
            "w",
        ) as fh:
            fh.write(yaml.dump(meta_yml))
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.nfcore_modules)
        module_lint.lint(print_results=False, module="bpipe/test")

        # reset changes
        meta_yml["name"] = "bpipe_test"
        with open(
            Path(self.nfcore_modules, "modules", "nf-core", "bpipe", "test", "meta.yml"),
            "w",
        ) as fh:
            fh.write(yaml.dump(meta_yml))

        assert len(module_lint.failed) == 1, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) >= 0
        assert len(module_lint.warned) >= 0
        assert module_lint.failed[0].lint_test == "meta_name"

    def test_modules_missing_test_dir(self):
        """Test linting a module with a missing test directory"""
        Path(self.nfcore_modules, "modules", "nf-core", "bpipe", "test", "tests").rename(
            Path(self.nfcore_modules, "modules", "nf-core", "bpipe", "test", "tests.bak")
        )
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.nfcore_modules)
        module_lint.lint(print_results=False, module="bpipe/test")
        Path(self.nfcore_modules, "modules", "nf-core", "bpipe", "test", "tests.bak").rename(
            Path(self.nfcore_modules, "modules", "nf-core", "bpipe", "test", "tests")
        )
        assert len(module_lint.failed) == 1, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) >= 0
        assert len(module_lint.warned) >= 0
        assert module_lint.failed[0].lint_test == "test_dir_exists"

    def test_modules_missing_test_main_nf(self):
        """Test linting a module with a missing test/main.nf file"""
        Path(
            self.nfcore_modules,
            "modules",
            "nf-core",
            "bpipe",
            "test",
            "tests",
            "main.nf.test",
        ).rename(
            Path(
                self.nfcore_modules,
                "modules",
                "nf-core",
                "bpipe",
                "test",
                "tests",
                "main.nf.test.bak",
            )
        )
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.nfcore_modules)
        module_lint.lint(print_results=False, module="bpipe/test")
        Path(
            self.nfcore_modules,
            "modules",
            "nf-core",
            "bpipe",
            "test",
            "tests",
            "main.nf.test.bak",
        ).rename(
            Path(
                self.nfcore_modules,
                "modules",
                "nf-core",
                "bpipe",
                "test",
                "tests",
                "main.nf.test",
            )
        )
        assert len(module_lint.failed) == 1, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) >= 0
        assert len(module_lint.warned) >= 0
        assert module_lint.failed[0].lint_test == "test_main_nf_exists"

    def test_modules_unused_pytest_files(self):
        """Test linting a nf-test module with files still present in `tests/modules/`"""
        Path(self.nfcore_modules, "tests", "modules", "bpipe", "test").mkdir(parents=True, exist_ok=True)
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.nfcore_modules)
        module_lint.lint(print_results=False, module="bpipe/test")
        Path(self.nfcore_modules, "tests", "modules", "bpipe", "test").rmdir()
        assert len(module_lint.failed) == 1, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) >= 0
        assert len(module_lint.warned) >= 0
        assert module_lint.failed[0].lint_test == "test_old_test_dir"

    def test_nftest_failing_linting(self):
        """Test linting a module which includes other modules in nf-test tests.
        Linting tests"""
        # Clone modules repo with testing modules
        tmp_dir = self.nfcore_modules.parent
        self.nfcore_modules = Path(tmp_dir, "modules-test")
        Repo.clone_from(GITLAB_URL, self.nfcore_modules, branch=GITLAB_NFTEST_BRANCH)

        module_lint = nf_core.modules.lint.ModuleLint(directory=self.nfcore_modules)
        module_lint.lint(print_results=False, module="kallisto/quant")

        assert len(module_lint.failed) == 3, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) >= 0
        assert len(module_lint.warned) >= 0
        assert module_lint.failed[0].lint_test == "environment_yml_valid"
        assert module_lint.failed[1].lint_test == "meta_yml_valid"
        assert module_lint.failed[2].lint_test == "test_main_tags"
        assert "kallisto/index" in module_lint.failed[2].message

    def test_modules_absent_version(self):
        """Test linting a nf-test module if the versions is absent in the snapshot file `"""
        with open(
            Path(
                self.nfcore_modules,
                "modules",
                "nf-core",
                "bpipe",
                "test",
                "tests",
                "main.nf.test.snap",
            )
        ) as fh:
            content = fh.read()
            new_content = content.replace("versions", "foo")
        with open(
            Path(
                self.nfcore_modules,
                "modules",
                "nf-core",
                "bpipe",
                "test",
                "tests",
                "main.nf.test.snap",
            ),
            "w",
        ) as fh:
            fh.write(new_content)
        module_lint = nf_core.modules.lint.ModuleLint(directory=self.nfcore_modules)
        module_lint.lint(print_results=False, module="bpipe/test")
        with open(
            Path(
                self.nfcore_modules,
                "modules",
                "nf-core",
                "bpipe",
                "test",
                "tests",
                "main.nf.test.snap",
            ),
            "w",
        ) as fh:
            fh.write(content)
        assert len(module_lint.failed) == 1, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) >= 0
        assert len(module_lint.warned) >= 0
        assert module_lint.failed[0].lint_test == "test_snap_versions"

    def test_modules_empty_file_in_snapshot(self):
        """Test linting a nf-test module with an empty file sha sum in the test snapshot, which should make it fail (if it is not a stub)"""
        snap_file = Path(
            self.nfcore_modules,
            "modules",
            "nf-core",
            "bpipe",
            "test",
            "tests",
            "main.nf.test.snap",
        )
        snap = json.load(snap_file.open())
        content = snap_file.read_text()
        snap["my test"]["content"][0]["0"] = "test:md5,d41d8cd98f00b204e9800998ecf8427e"

        with open(snap_file, "w") as fh:
            json.dump(snap, fh)

        module_lint = nf_core.modules.lint.ModuleLint(directory=self.nfcore_modules)
        module_lint.lint(print_results=False, module="bpipe/test")
        assert len(module_lint.failed) == 1, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) > 0
        assert len(module_lint.warned) >= 0
        assert module_lint.failed[0].lint_test == "test_snap_md5sum"

        # reset the file
        with open(snap_file, "w") as fh:
            fh.write(content)

    def test_modules_empty_file_in_stub_snapshot(self):
        """Test linting a nf-test module with an empty file sha sum in the stub test snapshot, which should make it not fail"""
        snap_file = Path(
            self.nfcore_modules,
            "modules",
            "nf-core",
            "bpipe",
            "test",
            "tests",
            "main.nf.test.snap",
        )
        snap = json.load(snap_file.open())
        content = snap_file.read_text()
        snap["my_test_stub"] = {"content": [{"0": "test:md5,d41d8cd98f00b204e9800998ecf8427e", "versions": {}}]}

        with open(snap_file, "w") as fh:
            json.dump(snap, fh)

        module_lint = nf_core.modules.lint.ModuleLint(directory=self.nfcore_modules)
        module_lint.lint(print_results=False, module="bpipe/test")
        assert len(module_lint.failed) == 0, f"Linting failed with {[x.__dict__ for x in module_lint.failed]}"
        assert len(module_lint.passed) > 0
        assert len(module_lint.warned) >= 0
        assert any(x.lint_test == "test_snap_md5sum" for x in module_lint.passed)

        # reset the file
        with open(snap_file, "w") as fh:
            fh.write(content)


# A skeleton object with the passed/warned/failed list attrs
# Use this in place of a ModuleLint object to test behaviour of
# linting methods which don't need the full setup
class MockModuleLint:
    def __init__(self):
        self.passed = []
        self.warned = []
        self.failed = []

        self.main_nf = "main_nf"



"""
Verify that a module has a correct entry in the modules.json file
"""

import logging
from pathlib import Path

import nf_core
import nf_core.modules.lint
import nf_core.modules.modules_repo
import nf_core.modules.modules_utils
from nf_core.modules.modules_utils import NFCoreComponent

log = logging.getLogger(__name__)


def module_version(module_lint_object: "nf_core.modules.lint.ModuleLint", module: "NFCoreComponent"):
    """
    Verifies that the module has a version specified in the ``modules.json`` file

    It checks whether the module has an entry in the ``modules.json`` file
    containing a commit SHA. If that is true, it verifies that there are no
    newer version of the module available.
    """
    assert module_lint_object.modules_json is not None  # mypy
    assert module.repo_url is not None  # mypy
    modules_json_path = Path(module_lint_object.directory, "modules.json")
    # Verify that a git_sha exists in the `modules.json` file for this module
    version = module_lint_object.modules_json.get_module_version(module.component_name, module.repo_url, module.org)
    if version is None:
        module.failed.append(("git_sha", "No git_sha entry in `modules.json`", modules_json_path))
        return

    module.git_sha = version
    module.passed.append(("git_sha", "Found git_sha entry in `modules.json`", modules_json_path))

    # Check whether a new version is available
    try:
        module.branch = module_lint_object.modules_json.get_component_branch(
            "modules", module.component_name, module.repo_url, module.org
        )
        modules_repo = nf_core.modules.modules_repo.ModulesRepo(remote_url=module.repo_url, branch=module.branch)

        module_git_log = list(modules_repo.get_component_git_log(module.component_name, "modules"))
        if version == module_git_log[0]["git_sha"]:
            module.passed.append(("module_version", "Module is the latest version", module.component_dir))
        else:
            module.warned.append(("module_version", "New version available", module.component_dir))
    except UserWarning:
        module.warned.append(("module_version", "Failed to fetch git log", module.component_dir))



"""Some tests covering the bump_version code."""

import yaml

import nf_core.pipelines.bump_version
import nf_core.utils

from ..test_pipelines import TestPipelines


class TestBumpVersion(TestPipelines):
    def test_bump_pipeline_version(self):
        """Test that making a release with the working example files works"""

        # Bump the version number
        nf_core.pipelines.bump_version.bump_pipeline_version(self.pipeline_obj, "1.1.0")
        new_pipeline_obj = nf_core.utils.Pipeline(self.pipeline_dir)

        # Check nextflow.config
        new_pipeline_obj.load_pipeline_config()
        assert new_pipeline_obj.nf_config["manifest.version"].strip("'\"") == "1.1.0"

        # Check multiqc_config.yml
        with open(new_pipeline_obj._fp("assets/multiqc_config.yml")) as fh:
            multiqc_config = yaml.safe_load(fh)

            assert "report_comment" in multiqc_config
            assert "/releases/tag/1.1.0" in multiqc_config["report_comment"]

        # Check .nf-core.yml
        with open(new_pipeline_obj._fp(".nf-core.yml")) as fh:
            nf_core_yml = yaml.safe_load(fh)
            if nf_core_yml["template"]:
                assert nf_core_yml["template"]["version"] == "1.1.0"

    def test_dev_bump_pipeline_version(self):
        """Test that making a release works with a dev name and a leading v"""
        # Bump the version number
        nf_core.pipelines.bump_version.bump_pipeline_version(self.pipeline_obj, "v1.2dev")
        new_pipeline_obj = nf_core.utils.Pipeline(self.pipeline_dir)

        # Check the pipeline config
        new_pipeline_obj.load_pipeline_config()
        assert new_pipeline_obj.nf_config["manifest.version"].strip("'\"") == "1.2dev"

    def test_bump_nextflow_version(self):
        # Bump the version number to a specific version, preferably one
        # we're not already on
        version = "25.04.2"
        nf_core.pipelines.bump_version.bump_nextflow_version(self.pipeline_obj, version)
        new_pipeline_obj = nf_core.utils.Pipeline(self.pipeline_dir)
        new_pipeline_obj._load()

        # Check nextflow.config
        assert new_pipeline_obj.nf_config["manifest.nextflowVersion"].strip("'\"") == f"!>={version}"

        # Check .github/workflows/ci.yml
        with open(new_pipeline_obj._fp(".github/workflows/ci.yml")) as fh:
            ci_yaml = yaml.safe_load(fh)
        assert ci_yaml["jobs"]["test"]["strategy"]["matrix"]["NXF_VER"][0] == version

        # Check README.md
        with open(new_pipeline_obj._fp("README.md")) as fh:
            readme = fh.read().splitlines()
        assert (
            f"[![Nextflow](https://img.shields.io/badge/nextflow%20DSL2-%E2%89%A5{version}-23aa62.svg)]"
            "(https://www.nextflow.io/)" in readme
        )



from pathlib import Path
from typing import Dict, List

from ruamel.yaml import YAML

from nf_core import __version__

REPOSITORY_TYPES = ["pipeline", "modules"]


def nfcore_yml(self) -> Dict[str, List[str]]:
    """Repository ``.nf-core.yml`` tests

    The ``.nf-core.yml`` contains metadata for nf-core tools to correctly apply its features.

    * repository type:

        * Check that the repository type is set.

    * nf core version:

         * Check if the nf-core version is set to the latest version.

    """
    passed: List[str] = []
    warned: List[str] = []
    failed: List[str] = []
    ignored: List[str] = []

    yaml = YAML()

    # Remove field that should be ignored according to the linting config
    ignore_configs = self.lint_config.get(".nf-core", []) if self.lint_config is not None else []
    for ext in (".yml", ".yaml"):
        try:
            nf_core_yml = yaml.load(Path(self.wf_path) / f".nf-core{ext}")
            break
        except FileNotFoundError:
            continue
    else:
        raise FileNotFoundError("No `.nf-core.yml` file found.")

    if "repository_type" not in ignore_configs:
        # Check that the repository type is set in the .nf-core.yml
        if "repository_type" in nf_core_yml:
            repo_type = nf_core_yml["repository_type"]
            if repo_type not in REPOSITORY_TYPES:
                failed.append(
                    f"Repository type in `.nf-core.yml` is not valid. "
                    f"Should be one of `[{', '.join(REPOSITORY_TYPES)}]` but was `{repo_type}`"
                )
            else:
                passed.append(f"Repository type in `.nf-core.yml` is valid: `{repo_type}`")
        else:
            warned.append("Repository type not set in `.nf-core.yml`")
    else:
        ignored.append("`.nf-core.yml` variable ignored 'repository_type'")

    if "nf_core_version" not in ignore_configs:
        # Check that the nf-core version is set in the .nf-core.yml
        if "nf_core_version" in nf_core_yml:
            nf_core_version = nf_core_yml["nf_core_version"]
            if nf_core_version != __version__ and "dev" not in nf_core_version:
                warned.append(
                    f"nf-core version in `.nf-core.yml` is not set to the latest version. "
                    f"Should be `{__version__}` but was `{nf_core_version}`"
                )
            else:
                passed.append(f"nf-core version in `.nf-core.yml` is set to the latest version: `{nf_core_version}`")
        else:
            warned.append("nf-core version not set in `.nf-core.yml`")
    else:
        ignored.append("`.nf-core.yml` variable ignored 'nf_core_version'")

    return {"passed": passed, "warned": warned, "failed": failed, "ignored": ignored}



import fnmatch
import logging
import os
from pathlib import Path

log = logging.getLogger(__name__)


def pipeline_todos(self, root_dir=None):
    """Check for nf-core *TODO* lines.

    The nf-core workflow template contains a number of comment lines to help developers
    of new pipelines know where they need to edit files and add content.
    They typically have the following format:

    .. code-block:: groovy

        // TODO nf-core: Make some kind of change to the workflow here

    ..or in markdown:

    .. code-block:: html

        <!-- TODO nf-core: Add some detail to the docs here -->

    This lint test runs through all files in the pipeline and searches for these lines.
    If any are found they will throw a warning.

    .. tip:: Note that many GUI code editors have plugins to list all instances of *TODO*
              in a given project directory. This is a very quick and convenient way to get
              started on your pipeline!
    """
    passed = []
    warned = []
    file_paths = []

    # Pipelines don't provide a path, so use the workflow path.
    # Modules run this function twice and provide a string path
    if root_dir is None:
        root_dir = self.wf_path

    ignore = [".git"]
    if Path(root_dir, ".gitignore").is_file():
        with open(Path(root_dir, ".gitignore"), encoding="latin1") as fh:
            for line in fh:
                ignore.append(Path(line.strip().rstrip("/")).name)
    for root, dirs, files in os.walk(root_dir, topdown=True):
        # Ignore files
        for i_base in ignore:
            i = str(Path(root, i_base))
            dirs[:] = [d for d in dirs if not fnmatch.fnmatch(str(Path(root, d)), i)]
            files[:] = [f for f in files if not fnmatch.fnmatch(str(Path(root, f)), i)]
        for fname in files:
            try:
                with open(Path(root, fname), encoding="latin1") as fh:
                    for line in fh:
                        if "TODO nf-core" in line:
                            line = (
                                line.replace("<!--", "")
                                .replace("-->", "")
                                .replace("# TODO nf-core: ", "")
                                .replace("// TODO nf-core: ", "")
                                .replace("TODO nf-core: ", "")
                                .strip()
                            )
                            warned.append(f"TODO string in `{fname}`: _{line}_")
                            file_paths.append(Path(root, fname))
            except FileNotFoundError:
                log.debug(f"Could not open file {fname} in pipeline_todos lint test")

    if len(warned) == 0:
        passed.append("No TODO strings found")

    # HACK file paths are returned to allow usage of this function in modules/lint.py
    # Needs to be refactored!
    return {"passed": passed, "warned": warned, "file_paths": file_paths}



from nf_core.components.update import ComponentUpdate


class ModuleUpdate(ComponentUpdate):
    def __init__(
        self,
        pipeline_dir,
        force=False,
        prompt=False,
        sha=None,
        update_all=False,
        show_diff=None,
        save_diff_fn=None,
        update_deps=False,
        remote_url=None,
        branch=None,
        no_pull=False,
        limit_output=False,
    ):
        super().__init__(
            pipeline_dir,
            "modules",
            force,
            prompt,
            sha,
            update_all,
            show_diff,
            save_diff_fn,
            update_deps,
            remote_url,
            branch,
            no_pull,
            limit_output,
        )



"""Code to deal with pipeline JSON Schema"""

import copy
import json
import logging
import tempfile
import webbrowser
from pathlib import Path
from typing import Union

import jinja2
import jsonschema
import markdown
import rich.console
import yaml
from rich.prompt import Confirm
from rich.syntax import Syntax

import nf_core.pipelines.list
import nf_core.utils
from nf_core.pipelines.lint_utils import dump_json_with_prettier, run_prettier_on_file

log = logging.getLogger(__name__)


class PipelineSchema:
    """Class to generate a schema object with
    functions to handle pipeline JSON Schema"""

    def __init__(self):
        """Initialise the object"""

        self.schema = {}
        self.pipeline_dir = ""
        self._schema_filename = ""
        self.schema_defaults = {}
        self.schema_types = {}
        self.schema_params = {}
        self.input_params = {}
        self.pipeline_params = {}
        self.invalid_nextflow_config_default_parameters = {}
        self.pipeline_manifest = {}
        self.schema_from_scratch = False
        self.no_prompts = False
        self.web_only = False
        self.web_schema_build_url = "https://oldsite.nf-co.re/pipeline_schema_builder"
        self.web_schema_build_web_url = None
        self.web_schema_build_api_url = None
        self.validation_plugin = None
        self.schema_draft = None
        self.defs_notation = None
        self.ignored_params = []

    # Update the validation plugin code every time the schema gets changed
    def set_schema_filename(self, schema: str) -> None:
        self._schema_filename = schema
        self._update_validation_plugin_from_config()

    def get_schema_filename(self) -> str:
        return self._schema_filename

    def del_schema_filename(self) -> None:
        del self._schema_filename

    schema_filename = property(get_schema_filename, set_schema_filename, del_schema_filename)

    def _update_validation_plugin_from_config(self) -> None:
        plugin = "nf-schema"
        if self.schema_filename:
            conf = nf_core.utils.fetch_wf_config(Path(self.schema_filename).parent)
        else:
            conf = nf_core.utils.fetch_wf_config(Path(self.pipeline_dir))

        plugins = str(conf.get("plugins", "")).strip("'\"").strip(" ").split(",")
        plugin_found = False
        for plugin_instance in plugins:
            if "nf-schema" in plugin_instance:
                plugin = "nf-schema"
                plugin_found = True
                break
            elif "nf-validation" in plugin_instance:
                plugin = "nf-validation"
                plugin_found = True
                break

        if not plugin_found:
            log.info(
                "Could not find nf-schema or nf-validation in the pipeline config. Defaulting to nf-schema notation for the JSON schema."
            )

        self.validation_plugin = plugin
        # Previous versions of nf-schema used "defs", but it's advised to use "$defs"
        if plugin == "nf-schema":
            self.defs_notation = "$defs"
            ignored_params = [
                conf.get("validation.help.shortParameter", "help"),
                conf.get("validation.help.fullParameter", "helpFull"),
                conf.get("validation.help.showHiddenParameter", "showHidden"),
                "trace_report_suffix",  # report suffix should be ignored by default as it is a Java Date object
            ]  # Help parameter should be ignored by default
            ignored_params_config_str = conf.get("validation.defaultIgnoreParams", "")
            ignored_params_config = [
                item.strip().strip("'") for item in ignored_params_config_str[1:-1].split(",")
            ]  # Extract list elements and remove whitespace

            if len(ignored_params_config) > 0:
                log.debug(f"Ignoring parameters from config: {ignored_params_config}")
                ignored_params.extend(ignored_params_config)
            self.ignored_params = ignored_params
            log.debug(f"Ignoring parameters: {self.ignored_params}")
            self.schema_draft = "https://json-schema.org/draft/2020-12/schema"

        else:
            self.defs_notation = "definitions"
            self.schema_draft = "https://json-schema.org/draft-07/schema"
            self.get_wf_params()
            self.ignored_params = self.pipeline_params.get("validationSchemaIgnoreParams", "").strip("\"'").split(",")
            self.ignored_params.append("validationSchemaIgnoreParams")

    def get_schema_path(
        self, path: Union[str, Path], local_only: bool = False, revision: Union[str, None] = None
    ) -> None:
        """Given a pipeline name, directory, or path, set self.schema_filename"""
        path = Path(path)
        # Supplied path exists - assume a local pipeline directory or schema
        if path.exists():
            log.debug(f"Path exists: {path}. Assuming local pipeline directory or schema")
            local_only = True
            if revision is not None:
                log.warning(f"Local workflow supplied, ignoring revision '{revision}'")
            if path.is_dir():
                self.pipeline_dir = path
                self.schema_filename = path / "nextflow_schema.json"
            else:
                self.pipeline_dir = path.parent
                self.schema_filename = path

        # Path does not exist - assume a name of a remote workflow
        elif not local_only:
            self.pipeline_dir = nf_core.pipelines.list.get_local_wf(path, revision=revision)
            self.schema_filename = Path(self.pipeline_dir or "", "nextflow_schema.json")
            # check if the schema file exists
            if not self.schema_filename.exists():
                self.schema_filename = None
        # Only looking for local paths, overwrite with None to be safe
        else:
            self.schema_filename = None

        # Check that the schema file exists
        if self.schema_filename is None or not Path(self.schema_filename).exists() and local_only:
            error = f"Could not find pipeline schema for '{path}': {self.schema_filename}"
            log.error(error)
            raise AssertionError(error)

    def load_lint_schema(self):
        """Load and lint a given schema to see if it looks valid"""
        try:
            self.load_schema()
            num_params = self.validate_schema()
            self.get_schema_defaults()
            self.validate_default_params()
            if len(self.invalid_nextflow_config_default_parameters) > 0:
                params = "\n --".join(
                    [f"{param}: {msg}" for param, msg in self.invalid_nextflow_config_default_parameters.items()]
                )
                log.info(
                    f"[red][✗] Invalid default parameters found:\n {params} \n\nNOTE: Use null in config for no default."
                )

            else:
                log.info(f"[green][✓] Pipeline schema looks valid[/] [dim](found {num_params} params)")
        except json.decoder.JSONDecodeError as e:
            error_msg = f"[bold red]Could not parse schema JSON:[/] {e}"
            log.error(error_msg)
            raise AssertionError(error_msg)
        except AssertionError as e:
            error_msg = f"[red][✗] Pipeline schema does not follow nf-core specs:\n {e}"
            log.error(error_msg)
            raise AssertionError(error_msg)

    def load_schema(self):
        """Load a pipeline schema from a file"""
        if self.schema_filename is None or not Path(self.schema_filename).exists():
            raise AssertionError("Pipeline schema filename could not be found.")

        with open(self.schema_filename) as fh:
            self.schema = json.load(fh)
        self.schema_defaults = {}
        self.schema_params = {}
        if "$schema" not in self.schema:
            raise AssertionError("Schema missing top-level `$schema` attribute")
        log.debug(f"JSON file loaded: {self.schema_filename}")

    def sanitise_param_default(self, param):
        """
        Given a param, ensure that the default value is the correct variable type
        """
        if "type" not in param or "default" not in param:
            return param

        # Bools
        if param["type"] == "boolean":
            if not isinstance(param["default"], bool):
                param["default"] = param["default"] == "true"
            return param

        # For everything else, an empty string is an empty string
        if isinstance(param["default"], str) and param["default"].strip() == "":
            param["default"] = ""
            return param

        # Integers
        if param["type"] == "integer":
            param["default"] = int(param["default"])
            return param

        # Numbers
        if param["type"] == "number":
            param["default"] = float(param["default"])
            return param

        if param["default"] is None:
            return param

        # Strings
        param["default"] = str(param["default"])
        return param

    def get_schema_defaults(self) -> None:
        """
        Generate set of default input parameters from schema.

        Saves defaults to self.schema_defaults
        Returns count of how many parameters were found (with or without a default value)
        """
        # Top level schema-properties (ungrouped)
        for p_key, param in self.schema.get("properties", {}).items():
            self.schema_params[p_key] = ("properties", p_key)
            if "default" in param:
                param = self.sanitise_param_default(param)
                if param["default"] is not None:
                    self.schema_defaults[p_key] = param["default"]

        # TODO add support for nested parameters
        # Grouped schema properties in subschema definitions
        for defn_name, definition in self.schema.get(self.defs_notation, {}).items():
            for p_key, param in definition.get("properties", {}).items():
                self.schema_params[p_key] = (self.defs_notation, defn_name, "properties", p_key)
                if "default" in param:
                    param = self.sanitise_param_default(param)
                    if param["default"] is not None:
                        self.schema_defaults[p_key] = param["default"]

    def get_schema_types(self) -> None:
        """Get a list of all parameter types in the schema"""
        for name, param in self.schema.get("properties", {}).items():
            if "type" in param:
                self.schema_types[name] = param["type"]
        for _, definition in self.schema.get(self.defs_notation, {}).items():
            for name, param in definition.get("properties", {}).items():
                if "type" in param:
                    self.schema_types[name] = param["type"]

    def save_schema(self, suppress_logging=False):
        """Save a pipeline schema to a file"""
        # Write results to a JSON file
        num_params = len(self.schema.get("properties", {}))
        num_params += sum(len(d.get("properties", {})) for d in self.schema.get(self.defs_notation, {}).values())
        if not suppress_logging:
            log.info(f"Writing schema with {num_params} params: '{self.schema_filename}'")
        dump_json_with_prettier(self.schema_filename, self.schema)

    def load_input_params(self, params_path):
        """Load a given a path to a parameters file (JSON/YAML)

        These should be input parameters used to run a pipeline with
        the Nextflow -params-file option.
        """
        # First, try to load as JSON
        try:
            with open(params_path) as fh:
                try:
                    params = json.load(fh)
                except json.JSONDecodeError as e:
                    raise UserWarning(f"Unable to load JSON file '{params_path}' due to error {e}")
                self.input_params.update(params)
            log.debug(f"Loaded JSON input params: {params_path}")
        except Exception as json_e:
            log.debug(f"Could not load input params as JSON: {json_e}")
            # This failed, try to load as YAML
            try:
                with open(params_path) as fh:
                    params = yaml.safe_load(fh)
                    self.input_params.update(params)
                    log.debug(f"Loaded YAML input params: {params_path}")
            except Exception as yaml_e:
                error_msg = f"Could not load params file as either JSON or YAML:\n JSON: {json_e}\n YAML: {yaml_e}"
                log.error(error_msg)
                raise AssertionError(error_msg)

    def validate_params(self):
        """Check given parameters against a schema and validate"""
        if self.schema is None:
            log.error("[red][✗] Pipeline schema not found")
            return False
        try:
            jsonschema.validate(self.input_params, self.schema)
        except jsonschema.exceptions.ValidationError as e:
            log.error(f"[red][✗] Input parameters are invalid: {e.message}")
            return False
        log.info("[green][✓] Input parameters look valid")
        return True

    def validate_default_params(self):
        """
        Check that all default parameters in the schema are valid
        Ignores 'required' flag, as required parameters might have no defaults

        Additional check that all parameters have defaults in nextflow.config and that
        these are valid and adhere to guidelines
        """
        if self.schema is None:
            log.error("[red][✗] Pipeline schema not found")
        try:
            # TODO add support for nested parameters
            # Make copy of schema and remove required flags
            schema_no_required = copy.deepcopy(self.schema)
            if "required" in schema_no_required:
                schema_no_required.pop("required")
            for group_key, group in schema_no_required.get(self.defs_notation, {}).items():
                if "required" in group:
                    schema_no_required[self.defs_notation][group_key].pop("required")
            jsonschema.validate(self.schema_defaults, schema_no_required)
        except jsonschema.exceptions.ValidationError as e:
            raise AssertionError(f"Default parameters are invalid: {e.message}")
        for param, default in self.schema_defaults.items():
            if default in ("null", "", None, "None") or default is False:
                log.warning(
                    f"[yellow][!] Default parameter '{param}' is empty, null, or False. It is advisable to remove the default from the schema"
                )
        log.info("[green][✓] Default parameters match schema validation")

        # Make sure every default parameter exists in the nextflow.config and is of correct type
        if self.pipeline_params == {}:
            self.get_wf_params()

        # Go over group keys
        for group_key, group in schema_no_required.get(self.defs_notation, {}).items():
            group_properties = group.get("properties")
            for param in group_properties:
                if param in self.ignored_params:
                    continue
                if param in self.pipeline_params:
                    self.validate_config_default_parameter(param, group_properties[param], self.pipeline_params[param])
                else:
                    self.invalid_nextflow_config_default_parameters[param] = (
                        "Not in pipeline parameters. Check `nextflow.config`."
                    )

        # Go over ungrouped params if any exist
        ungrouped_properties = self.schema.get("properties")
        if ungrouped_properties:
            for param in ungrouped_properties:
                if param in self.ignored_params:
                    continue
                if param in self.pipeline_params:
                    self.validate_config_default_parameter(
                        param, ungrouped_properties[param], self.pipeline_params[param]
                    )
                else:
                    self.invalid_nextflow_config_default_parameters[param] = (
                        "Not in pipeline parameters. Check `nextflow.config`."
                    )

    def validate_config_default_parameter(self, param, schema_param, config_default):
        """
        Assure that default parameters in the nextflow.config are correctly set
        by comparing them to their type in the schema
        """

        # If we have a default in the schema, check it matches the config
        if "default" in schema_param and (
            (schema_param["type"] == "boolean" and str(config_default).lower() != str(schema_param["default"]).lower())
            and (str(schema_param["default"]) != str(config_default).strip("'\""))
        ):
            # Check that we are not deferring the execution of this parameter in the schema default with squiggly brakcets
            if schema_param["type"] != "string" or "{" not in schema_param["default"]:
                self.invalid_nextflow_config_default_parameters[param] = (
                    f"Schema default (`{schema_param['default']}`) does not match the config default (`{config_default}`)"
                )
                return

        # if default is null, we're good
        if config_default == "null":
            return

        # Check variable types in nextflow.config
        if schema_param["type"] == "string":
            if str(config_default) in ["false", "true", "''"]:
                self.invalid_nextflow_config_default_parameters[param] = (
                    f"String should not be set to `{config_default}`"
                )
        if schema_param["type"] == "boolean":
            if str(config_default) not in ["false", "true"]:
                self.invalid_nextflow_config_default_parameters[param] = (
                    f"Booleans should only be true or false, not `{config_default}`"
                )
        if schema_param["type"] == "integer":
            try:
                int(config_default)
            except ValueError:
                self.invalid_nextflow_config_default_parameters[param] = (
                    f"Does not look like an integer: `{config_default}`"
                )
        if schema_param["type"] == "number":
            try:
                float(config_default)
            except ValueError:
                self.invalid_nextflow_config_default_parameters[param] = (
                    f"Does not look like a number (float): `{config_default}`"
                )

    def validate_schema(self, schema=None):
        """
        Check that the Schema is valid

        Returns: Number of parameters found
        """
        if schema is None:
            schema = self.schema

        if "$schema" not in schema:
            raise AssertionError("Schema missing top-level `$schema` attribute")
        schema_draft = schema["$schema"]
        if self.schema_draft != schema_draft:
            raise AssertionError(f"Schema is using the wrong draft: {schema_draft}, should be {self.schema_draft}")
        if self.schema_draft == "https://json-schema.org/draft-07/schema":
            try:
                jsonschema.Draft7Validator.check_schema(schema)
                log.debug("JSON Schema Draft7 validated")
            except jsonschema.exceptions.SchemaError as e:
                raise AssertionError(f"Schema does not validate as Draft 7 JSON Schema:\n {e}")
        elif self.schema_draft == "https://json-schema.org/draft/2020-12/schema":
            try:
                jsonschema.Draft202012Validator.check_schema(schema)
                log.debug("JSON Schema Draft2020-12 validated")
            except jsonschema.exceptions.SchemaError as e:
                raise AssertionError(f"Schema does not validate as Draft 2020-12 JSON Schema:\n {e}")
        else:
            raise AssertionError(
                f"Schema `$schema` should be `https://json-schema.org/draft/2020-12/schema` or `https://json-schema.org/draft-07/schema` \n Found `{schema_draft}`"
            )

        param_keys = list(schema.get("properties", {}).keys())
        num_params = len(param_keys)

        # Add a small check for older nf-schema JSON schemas
        if "defs" in schema:
            raise AssertionError(
                f'Using "defs" for schema definitions is not supported. Please use "{self.defs_notation}" instead'
            )

        for d_key, d_schema in schema.get(self.defs_notation, {}).items():
            # Check that this definition is mentioned in allOf
            if "allOf" not in schema:
                raise AssertionError("Schema has definitions, but no allOf key")
            in_allOf = False
            for allOf in schema.get("allOf", []):
                if allOf["$ref"] == f"#/{self.defs_notation}/{d_key}":
                    in_allOf = True
            if not in_allOf:
                raise AssertionError(
                    f"Definition subschema `#/{self.defs_notation}/{d_key}` not included in schema `allOf`"
                )

            # TODO add support for nested parameters
            for d_param_id in d_schema.get("properties", {}):
                # Check that we don't have any duplicate parameter IDs in different definitions
                if d_param_id in param_keys:
                    raise AssertionError(f"Duplicate parameter found in schema `{self.defs_notation}`: `{d_param_id}`")
                param_keys.append(d_param_id)
                num_params += 1

        # Check that everything in allOf exists
        for allOf in schema.get("allOf", []):
            _, allof_defs_notation, def_key = allOf["$ref"].split("/")  # "#/<defs_notation>/<def_name>"
            if allof_defs_notation not in schema:
                raise AssertionError(f"Schema has allOf, but no {allof_defs_notation}")
            if def_key not in schema.get(allof_defs_notation, {}):
                raise AssertionError(f"Subschema `{def_key}` found in `allOf` but not `{allof_defs_notation}`")

        # Check that the schema describes at least one parameter
        if num_params == 0:
            raise AssertionError("No parameters found in schema")

        return num_params

    def validate_schema_title_description(self, schema=None):
        """
        Extra validation command for linting.
        Checks that the schema "$id", "title" and "description" attributes match the pipeline config.
        """
        if schema is None:
            schema = self.schema
        if schema is None:
            log.debug("Pipeline schema not set - skipping validation of top-level attributes")
            return None

        if self.pipeline_manifest == {}:
            self.get_wf_params()

        if "name" not in self.pipeline_manifest:
            log.debug("Pipeline manifest `name` not known - skipping validation of schema id and title")
        else:
            if "$id" not in self.schema:
                raise AssertionError("Schema missing top-level `$id` attribute")
            if "title" not in self.schema:
                raise AssertionError("Schema missing top-level `title` attribute")
            # Validate that id, title and description match the pipeline manifest
            id_attr = "https://raw.githubusercontent.com/{}/main/nextflow_schema.json".format(
                self.pipeline_manifest["name"].strip("\"'")
            )
            if self.schema["$id"] not in [id_attr, id_attr.replace("/main/", "/master/")]:
                raise AssertionError(
                    f"Schema `$id` should be `{id_attr}` or {id_attr.replace('/main/', '/master/')}. \n Found `{self.schema['$id']}`"
                )

            title_attr = "{} pipeline parameters".format(self.pipeline_manifest["name"].strip("\"'"))
            if self.schema["title"] != title_attr:
                raise AssertionError(f"Schema `title` should be `{title_attr}`\n Found: `{self.schema['title']}`")

        if "description" not in self.pipeline_manifest:
            log.debug("Pipeline manifest 'description' not known - skipping validation of schema description")
        else:
            if "description" not in self.schema:
                raise AssertionError("Schema missing top-level 'description' attribute")
            desc_attr = self.pipeline_manifest["description"].strip("\"'")
            if self.schema["description"] != desc_attr:
                raise AssertionError(
                    f"Schema 'description' should be '{desc_attr}'\n Found: '{self.schema['description']}'"
                )

    def check_for_input_mimetype(self):
        """
        Check that the input parameter has a mimetype

        Common mime types: https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/MIME_types/Common_types

        Returns:
            mimetype (str): The mimetype of the input parameter

        Raises:
            LookupError: If the input parameter is not found or defined in the correct place
        """

        # Check that the input parameter is defined
        if "input" not in self.schema_params:
            raise LookupError("Parameter `input` not found in schema")
        # Check that the input parameter is defined in the right place
        if "input" not in self.schema.get(self.defs_notation, {}).get("input_output_options", {}).get("properties", {}):
            raise LookupError("Parameter `input` is not defined in the correct subschema (input_output_options)")
        input_entry = self.schema[self.defs_notation]["input_output_options"]["properties"]["input"]
        if "mimetype" not in input_entry:
            return None
        mimetype = input_entry["mimetype"]
        if mimetype == "" or mimetype is None:
            return None
        return mimetype

    def print_documentation(
        self,
        output_fn=None,
        format="markdown",
        force=False,
        columns=None,
    ):
        """
        Prints documentation for the schema.
        """
        if columns is None:
            columns = ["parameter", "description", "type,", "default", "required", "hidden"]

        output = self.schema_to_markdown(columns)
        if format == "html":
            output = self.markdown_to_html(output)

        with tempfile.NamedTemporaryFile(mode="w+") as fh:
            fh.write(output)
            run_prettier_on_file(fh.name)
            fh.seek(0)
            prettified_docs = fh.read()

        if not output_fn:
            console = rich.console.Console()
            console.print("\n", Syntax(prettified_docs, format, word_wrap=True), "\n")
        else:
            if Path(output_fn).exists() and not force:
                log.error(f"File '{output_fn}' exists! Please delete first, or use '--force'")
                return
            with open(output_fn, "w") as fh:
                fh.write(prettified_docs)
                log.info(f"Documentation written to '{output_fn}'")

        # Return as a string
        return output

    def schema_to_markdown(self, columns):
        """
        Creates documentation for the schema in Markdown format.
        """
        out = f"# {self.schema['title']}\n\n"
        out += f"{self.schema['description']}\n"
        # Grouped parameters
        for definition in self.schema.get(self.defs_notation, {}).values():
            out += f"\n## {definition.get('title', {})}\n\n"
            out += f"{definition.get('description', '')}\n\n"
            required = definition.get("required", [])
            properties = definition.get("properties", {})
            param_table = self.markdown_param_table(properties, required, columns)
            out += param_table

        # Top-level ungrouped parameters
        if len(self.schema.get("properties", {})) > 0:
            out += "\n## Other parameters\n\n"
            required = self.schema.get("required", [])
            properties = self.schema.get("properties", {})
            param_table = self.markdown_param_table(properties, required, columns)
            out += param_table

        return out

    def markdown_param_table(self, properties, required, columns):
        """Creates a markdown table for params from jsonschema properties section

        Args:
            properties (dict): A jsonschema properties dictionary
            required (list): A list of the required fields.
                Should come from the same level of the jsonschema as properties
            columns (list): A list of columns to write

        Returns:
            str: A string with the markdown table
        """
        out = ""
        out += "".join([f"| {column.title()} " for column in columns])
        out += "|\n"
        out += "".join(["|-----------" for _ in columns])
        out += "|\n"
        for p_key, param in properties.items():
            for column in columns:
                if column == "parameter":
                    out += f"| `{p_key}` "
                elif column == "description":
                    desc = param.get("description", "").replace("\n", "<br>")
                    out += f"| {desc} "
                    if param.get("help_text", "") != "":
                        help_txt = param["help_text"].replace("\n", "<br>")
                        out += f"<details><summary>Help</summary><small>{help_txt}</small></details>"
                elif column == "type":
                    out += f"| `{param.get('type', '')}` "
                elif column == "required":
                    out += f"| {p_key in required or ''} "
                else:
                    out += f"| {param.get(column, '')} "
            out += "|\n"
        return out

    def markdown_to_html(self, markdown_str):
        """
        Convert markdown to html
        """
        return markdown.markdown(markdown_str, extensions=["tables"])

    def make_skeleton_schema(self):
        """Make a new pipeline schema from the template"""
        self.schema_from_scratch = True
        # Use Jinja to render the template schema file to a variable
        env = jinja2.Environment(
            loader=jinja2.PackageLoader("nf_core", "pipeline-template"), keep_trailing_newline=True
        )
        schema_template = env.get_template("nextflow_schema.json")

        template_vars = {
            "name": self.pipeline_manifest.get("name", Path(self.schema_filename).parent.name).strip("'"),
            "description": self.pipeline_manifest.get("description", "").strip("'"),
        }

        self.schema = json.loads(schema_template.render(template_vars))
        self.get_schema_defaults()

    def build_schema(self, pipeline_dir, no_prompts, web_only, url):
        """Interactively build a new pipeline schema for a pipeline"""

        # Check if supplied pipeline directory really is one
        nf_core.utils.is_pipeline_directory(pipeline_dir)

        if no_prompts:
            self.no_prompts = True
        if web_only:
            self.web_only = True
        if url:
            self.web_schema_build_url = url

        # Get pipeline schema filename
        try:
            self.get_schema_path(pipeline_dir, local_only=True)
        except AssertionError:
            log.info("No existing schema found - creating a new one from the nf-core template")
            self.get_wf_params()
            self.make_skeleton_schema()
            self.remove_schema_notfound_configs()
            self.remove_schema_empty_definitions()
            self.add_schema_found_configs()
            try:
                self.validate_schema()
            except AssertionError as e:
                log.error(f"[red]Something went wrong when building a new schema:[/] {e}")
                log.info("Please ask for help on the nf-core Slack")
                return False
        else:
            # Schema found - load and validate
            try:
                self.load_lint_schema()
            except AssertionError:
                log.error(f"Existing pipeline schema found, but it is invalid: {self.schema_filename}")
                log.info("Please fix or delete this file, then try again.")
                return False

        if not self.web_only:
            self.get_wf_params()
            self.remove_schema_notfound_configs()
            self.remove_schema_empty_definitions()
            self.add_schema_found_configs()
            self.save_schema()

        # If running interactively, send to the web for customisation
        if not self.no_prompts:
            if Confirm.ask(":rocket:  Launch web builder for customisation and editing?"):
                try:
                    self.launch_web_builder()
                except AssertionError as e:
                    log.error(e.args[0])
                    # Extra help for people running offline
                    if "Could not connect" in e.args[0]:
                        log.info(
                            f"If you're working offline, now copy your schema ({self.schema_filename}) and paste at https://nf-co.re/pipeline_schema_builder"
                        )
                        log.info("When you're finished, you can paste the edited schema back into the same file")
                    if self.web_schema_build_web_url:
                        log.info(
                            "To save your work, open {}\n"
                            f"Click the blue 'Finished' button, copy the schema and paste into this file: { self.web_schema_build_web_url, self.schema_filename}"
                        )
                    return False

    def get_wf_params(self):
        """
        Load the pipeline parameter defaults using `nextflow config`
        Strip out only the params. values and ignore anything that is not a flat variable
        """
        # Check that we haven't already pulled these (eg. skeleton schema)
        if len(self.pipeline_params) > 0 and len(self.pipeline_manifest) > 0:
            log.debug("Skipping get_wf_params as we already have them")
            return
        if self.schema_filename is None:
            log.error("Cannot get workflow params without a schema file")
            return
        log.debug("Collecting pipeline parameter defaults\n")
        config = nf_core.utils.fetch_wf_config(Path(self.schema_filename).parent)
        skipped_params = []
        # Pull out just the params. values
        for ckey, cval in config.items():
            if ckey.startswith("params."):
                # skip anything that's not a flat variable
                if "." in ckey[7:]:
                    skipped_params.append(ckey)
                    continue
                self.pipeline_params[ckey[7:]] = cval
            if ckey.startswith("manifest."):
                self.pipeline_manifest[ckey[9:]] = cval
        # Log skipped params
        if len(skipped_params) > 0:
            log.debug(
                "Skipped following pipeline params because they had nested parameter values:\n{}".format(
                    ", ".join(skipped_params)
                )
            )

    def remove_schema_empty_definitions(self):
        """
        Go through top-level schema remove definitions that don't have
        any property attributes
        """
        # Identify and remove empty definitions from the schema
        empty_definitions = []
        for d_key, d_schema in list(self.schema.get(self.defs_notation, {}).items()):
            if not d_schema.get("properties"):
                del self.schema[self.defs_notation][d_key]
                empty_definitions.append(d_key)
                log.warning(f"Removing empty group: '{d_key}'")

        # Remove "allOf" group with empty definitions from the schema
        for d_key in empty_definitions:
            allOf = {"$ref": f"#/{self.defs_notation}/{d_key}"}
            if allOf in self.schema.get("allOf", []):
                self.schema["allOf"].remove(allOf)

        # If we don't have anything left in "allOf", remove it
        if self.schema.get("allOf") == []:
            del self.schema["allOf"]

        # If we don't have anything left in "definitions", remove it
        if self.schema.get(self.defs_notation) == {}:
            del self.schema[self.defs_notation]

    def remove_schema_notfound_configs(self):
        """
        Go through top-level schema and all definitions sub-schemas to remove
        anything that's not in the nextflow config.
        """
        # Top-level properties
        self.schema, params_removed = self.remove_schema_notfound_configs_single_schema(self.schema)
        # Sub-schemas in definitions
        for d_key, definition in self.schema.get(self.defs_notation, {}).items():
            cleaned_schema, p_removed = self.remove_schema_notfound_configs_single_schema(definition)
            self.schema[self.defs_notation][d_key] = cleaned_schema
            params_removed.extend(p_removed)

        return params_removed

    def remove_schema_notfound_configs_single_schema(self, schema):
        """
        Go through a single schema / set of properties and strip out
        anything that's not in the nextflow config.

        Takes: Schema or sub-schema with properties key
        Returns: Cleaned schema / sub-schema
        """
        # Make a deep copy so as not to edit in place
        schema = copy.deepcopy(schema)
        params_removed = []
        # Use iterator so that we can delete the key whilst iterating
        for p_key in [k for k in schema.get("properties", {}).keys()]:
            if self.prompt_remove_schema_notfound_config(p_key):
                del schema["properties"][p_key]
                # Remove required flag if set
                if p_key in schema.get("required", []):
                    schema["required"].remove(p_key)
                # Remove required list if now empty
                if "required" in schema and len(schema["required"]) == 0:
                    del schema["required"]
                log.debug(f"Removing '{p_key}' from pipeline schema")
                params_removed.append(p_key)

        return schema, params_removed

    def prompt_remove_schema_notfound_config(self, p_key):
        """
        Check if a given key is found in the nextflow config params and prompt to remove it if note

        Returns True if it should be removed, False if not.
        """
        if p_key not in self.pipeline_params:
            if self.no_prompts or self.schema_from_scratch:
                return True
            if Confirm.ask(
                f":question: Unrecognised [bold]'params.{p_key}'[/] found in the schema but not in the pipeline config! [yellow]Remove it?"
            ):
                return True
        return False

    def add_schema_found_configs(self):
        """
        Add anything that's found in the Nextflow params that's missing in the pipeline schema
        Update defaults if they have changed
        """
        params_added = []

        for p_key, p_val in self.pipeline_params.items():
            s_key = self.schema_params.get(p_key)
            # Check if key is in schema parameters
            # Key is in pipeline but not in schema or ignored from schema
            if p_key not in self.schema_params and p_key not in self.ignored_params:
                if (
                    self.no_prompts
                    or self.schema_from_scratch
                    or Confirm.ask(
                        f":sparkles: Found [bold]'params.{p_key}'[/] in the pipeline config, but not in the schema. "
                        "[blue]Add to pipeline schema?"
                    )
                ):
                    if "properties" not in self.schema:
                        self.schema["properties"] = {}
                    self.schema["properties"][p_key] = self.build_schema_param(p_val)
                    log.debug(f"Adding '{p_key}' to pipeline schema")
                    params_added.append(p_key)
            # Param has a default that does not match the schema
            elif p_key in self.schema_defaults and (s_def := self.schema_defaults[p_key]) != (
                p_def := self.build_schema_param(p_val).get("default")
            ):
                if self.no_prompts or Confirm.ask(
                    f":sparkles: Default for [bold]'params.{p_key}'[/] in the pipeline config does not match schema. (schema: '{type(s_def)}: {s_def}'  | config: '{type(p_def)}: {p_def}'). "
                    "[blue]Update pipeline schema?"
                ):
                    s_key_def = s_key + ("default",)
                    if p_def is None:
                        nf_core.utils.nested_delitem(self.schema, s_key_def)
                        log.debug(f"Removed '{p_key}' default from pipeline schema")
                    else:
                        nf_core.utils.nested_setitem(self.schema, s_key_def, p_def)
                        log.debug(f"Updating '{p_key}' default to '{p_def}' in pipeline schema")
            # There is no default in schema but now there is a default to write
            elif (
                s_key
                and (p_key not in self.schema_defaults)
                and (p_key not in self.ignored_params)
                and (p_def := self.build_schema_param(p_val).get("default"))
            ):
                if self.no_prompts or Confirm.ask(
                    f":sparkles: Default for [bold]'params.{p_key}'[/] is not in schema (def='{p_def}'). "
                    "[blue]Update pipeline schema?"
                ):
                    s_key_def = s_key + ("default",)
                    nf_core.utils.nested_setitem(self.schema, s_key_def, p_def)
                    log.debug(f"Updating '{p_key}' default to '{p_def}' in pipeline schema")
        return params_added

    def build_schema_param(self, p_val):
        """
        Build a pipeline schema dictionary for an param interactively
        """
        p_val = p_val.strip("\"'")
        # p_val is always a string as it is parsed from nextflow config this way
        try:
            p_val = float(p_val)
            if p_val == int(p_val):
                p_val = int(p_val)
                p_type = "integer"
            else:
                p_type = "number"
        except ValueError:
            p_type = "string"

        # Anything can be "null", means that it is not set
        if p_val == "null":
            p_val = None

        # Booleans
        if p_val in ["true", "false", "True", "False"]:
            p_val = p_val in ["true", "True"]  # Convert to bool
            p_type = "boolean"

        # Don't return a default for anything false-y except 0
        if not p_val and not (p_val == 0 and p_val is not False):
            return {"type": p_type}

        return {"type": p_type, "default": p_val}

    def launch_web_builder(self):
        """
        Send pipeline schema to web builder and wait for response
        """

        content = {
            "post_content": "json_schema",
            "api": "true",
            "version": nf_core.__version__,
            "status": "waiting_for_user",
            "schema": json.dumps(self.schema),
        }
        web_response = nf_core.utils.poll_nfcore_web_api(self.web_schema_build_url, content)

        try:
            if "api_url" not in web_response:
                raise AssertionError('"api_url" not in web_response')
            if "web_url" not in web_response:
                raise AssertionError('"web_url" not in web_response')
            # DO NOT FIX THIS TYPO. Needs to stay in sync with the website. Maintaining for backwards compatibility.
            if web_response["status"] != "recieved":
                raise AssertionError(
                    f'web_response["status"] should be "recieved", but it is "{web_response["status"]}"'
                )
        except AssertionError:
            log.debug(f"Response content:\n{json.dumps(web_response, indent=4)}")
            raise AssertionError(
                f"Pipeline schema builder response not recognised: {self.web_schema_build_url}\n"
                " See verbose log for full response (nf-core -v schema)"
            )
        else:
            self.web_schema_build_web_url = web_response["web_url"]
            self.web_schema_build_api_url = web_response["api_url"]
            log.info(f"Opening URL: {web_response['web_url']}")
            webbrowser.open(web_response["web_url"])
            log.info("Waiting for form to be completed in the browser. Remember to click Finished when you're done.\n")
            nf_core.utils.wait_cli_function(self.get_web_builder_response)

    def get_web_builder_response(self):
        """
        Given a URL for a Schema build response, recursively query it until results are ready.
        Once ready, validate Schema and write to disk.
        """
        web_response = nf_core.utils.poll_nfcore_web_api(self.web_schema_build_api_url)
        if web_response["status"] == "error":
            raise AssertionError(f"Got error from schema builder: '{web_response.get('message')}'")
        if web_response["status"] == "waiting_for_user":
            return False
        if web_response["status"] == "web_builder_edited":
            log.info("Found saved status from nf-core pipelines schema builder")
            try:
                self.schema = web_response["schema"]
                self.remove_schema_empty_definitions()
                self.validate_schema()
            except AssertionError as e:
                raise AssertionError(f"Response from schema builder did not pass validation:\n {e}")
            else:
                self.save_schema()
                return True
        else:
            log.debug(f"Response content:\n{json.dumps(web_response, indent=4)}")
            raise AssertionError(
                f"Pipeline schema builder returned unexpected status ({web_response['status']}): "
                f"{self.web_schema_build_api_url}\n See verbose log for full response"
            )



import os
from pathlib import Path

from ..test_modules import TestModules


class TestModulesRemove(TestModules):
    def test_modules_remove_trimgalore(self):
        """Test removing TrimGalore! module after installing it"""
        self.mods_install.install("trimgalore")
        assert self.mods_install.directory is not None
        module_path = Path(self.mods_install.directory, "modules", "nf-core", "modules", "trimgalore")
        assert self.mods_remove.remove("trimgalore")
        assert os.path.exists(module_path) is False

    def test_modules_remove_trimgalore_uninstalled(self):
        """Test removing TrimGalore! module without installing it"""
        assert self.mods_remove.remove("trimgalore") is False

    def test_modules_remove_multiqc_from_gitlab(self):
        """Test removing multiqc module after installing it from an alternative source"""
        self.mods_install_gitlab.install("multiqc")
        assert self.mods_install.directory is not None
        module_path = Path(self.mods_install_gitlab.directory, "modules", "nf-core-test", "multiqc")
        assert self.mods_remove_gitlab.remove("multiqc", force=True)
        assert os.path.exists(module_path) is False



import os
import re
from pathlib import Path

import yaml

import nf_core.pipelines.create.create
import nf_core.pipelines.lint

from ..test_lint import TestLint


class TestLintNextflowConfig(TestLint):
    def setUp(self) -> None:
        super().setUp()
        self.new_pipeline = self._make_pipeline_copy()

    def test_nextflow_config_example_pass(self):
        """Tests that config variable existence test works with good pipeline example"""
        self.lint_obj.load_pipeline_config()
        result = self.lint_obj.nextflow_config()
        assert len(result["failed"]) == 0
        assert len(result["warned"]) == 0

    def test_default_values_match(self):
        """Test that the default values in nextflow.config match the default values defined in the nextflow_schema.json."""
        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj.load_pipeline_config()
        result = lint_obj.nextflow_config()
        assert len(result["failed"]) == 0
        assert len(result["warned"]) == 0
        assert "Config default value correct: params.validate_params" in str(result["passed"])

    def test_nextflow_config_bad_name_fail(self):
        """Tests that config variable existence test fails with bad pipeline name"""
        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj.load_pipeline_config()

        lint_obj.nf_config["manifest.name"] = "bad_name"
        result = lint_obj.nextflow_config()
        assert len(result["failed"]) > 0
        assert len(result["warned"]) == 0

    def test_nextflow_config_dev_in_release_mode_failed(self):
        """Tests that config variable existence test fails with dev version in release mode"""
        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj.load_pipeline_config()

        lint_obj.release_mode = True
        lint_obj.nf_config["manifest.version"] = "dev_is_bad_name"
        result = lint_obj.nextflow_config()
        assert len(result["failed"]) > 0
        assert len(result["warned"]) == 0

    def test_nextflow_config_missing_test_profile_failed(self):
        """Test failure if config file does not contain `test` profile."""
        # Change the name of the test profile so there is no such profile
        nf_conf_file = os.path.join(self.new_pipeline, "nextflow.config")
        with open(nf_conf_file) as f:
            content = f.read()
            fail_content = re.sub(r"\btest\b", "testfail", content)
        with open(nf_conf_file, "w") as f:
            f.write(fail_content)
        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj.load_pipeline_config()
        result = lint_obj.nextflow_config()
        assert len(result["failed"]) > 0
        assert len(result["warned"]) == 0

    def test_default_values_fail(self):
        """Test linting fails if the default values in nextflow.config do not match the ones defined in the nextflow_schema.json."""
        # Change the default value of max_multiqc_email_size in nextflow.config
        nf_conf_file = Path(self.new_pipeline) / "nextflow.config"
        with open(nf_conf_file) as f:
            content = f.read()
            fail_content = re.sub(r"\bmax_multiqc_email_size\s*=\s*'25.MB'", "max_multiqc_email_size = '0'", content)
        with open(nf_conf_file, "w") as f:
            f.write(fail_content)
        # Change the default value of custom_config_version in nextflow_schema.json
        nf_schema_file = Path(self.new_pipeline) / "nextflow_schema.json"
        with open(nf_schema_file) as f:
            content = f.read()
            fail_content = re.sub(r'"default": "master"', '"default": "main"', content)
        with open(nf_schema_file, "w") as f:
            f.write(fail_content)
        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj.load_pipeline_config()
        result = lint_obj.nextflow_config()
        assert len(result["failed"]) == 2
        assert (
            "Config default value incorrect: `params.max_multiqc_email_size` is set as `25.MB` in `nextflow_schema.json` but is `0` in `nextflow.config`."
            in result["failed"]
        )
        assert (
            "Config default value incorrect: `params.custom_config_version` is set as `main` in `nextflow_schema.json` but is `master` in `nextflow.config`."
            in result["failed"]
        )

    def test_catch_params_assignment_in_main_nf(self):
        """Test linting fails if main.nf contains an assignment to a parameter from nextflow_schema.json."""
        # Add parameter assignment in main.nf
        main_nf_file = Path(self.new_pipeline) / "main.nf"
        with open(main_nf_file, "a") as f:
            f.write("params.custom_config_base = 'test'")
        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj.load_pipeline_config()
        result = lint_obj.nextflow_config()
        assert len(result["failed"]) == 2
        assert (
            result["failed"][1]
            == "Config default value incorrect: `params.custom_config_base` is set as `https://raw.githubusercontent.com/nf-core/configs/master` in `nextflow_schema.json` but is `null` in `nextflow.config`."
        )

    def test_allow_params_reference_in_main_nf(self):
        """Test linting allows for references like `params.aligner == 'bwa'` in main.nf. The test will detect if the bug mentioned in GitHub-issue #2833 reemerges."""
        # Add parameter reference in main.nf
        main_nf_file = Path(self.new_pipeline) / "main.nf"
        with open(main_nf_file, "a") as f:
            f.write("params.custom_config_version == 'main'")
        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj.load_pipeline_config()
        result = lint_obj.nextflow_config()
        assert len(result["failed"]) == 0

    def test_default_values_ignored(self):
        """Test ignoring linting of default values."""
        valid_yaml = """
        nextflow_config:
            - manifest.name
            - config_defaults:
                - params.custom_config_version
        """
        # Add custom_config_version to the ignore list
        nf_core_yml_path = Path(self.new_pipeline) / ".nf-core.yml"

        with open(nf_core_yml_path) as f:
            nf_core_yml = yaml.safe_load(f)
            nf_core_yml["lint"] = yaml.safe_load(valid_yaml)
        with open(nf_core_yml_path, "w") as f:
            yaml.dump(nf_core_yml, f)

        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj.load_pipeline_config()
        lint_obj._load_lint_config()
        result = lint_obj.nextflow_config()
        assert len(result["failed"]) == 0
        assert len(result["ignored"]) == 2
        assert "Config default value correct: params.custom_config_version" not in str(result["passed"])
        assert "Config default ignored: params.custom_config_version" in str(result["ignored"])
        assert "Config variable ignored: `manifest.name`" in str(result["ignored"])

    def test_default_values_float(self):
        """Test comparing two float values."""
        # Add a float value `dummy=0.0001` to the nextflow.config below `validate_params`
        nf_conf_file = Path(self.new_pipeline) / "nextflow.config"
        with open(nf_conf_file) as f:
            content = f.read()
            fail_content = re.sub(
                r"validate_params\s*=\s*true",
                "params.validate_params = true\ndummy = 0.000000001",
                content,
            )
        with open(nf_conf_file, "w") as f:
            f.write(fail_content)
        # Add a float value `dummy` to the nextflow_schema.json
        nf_schema_file = Path(self.new_pipeline) / "nextflow_schema.json"
        with open(nf_schema_file) as f:
            content = f.read()
            fail_content = re.sub(
                r'"validate_params": {',
                '    "dummy": {"type": "number","default":0.000000001},\n"validate_params": {',
                content,
            )
        with open(nf_schema_file, "w") as f:
            f.write(fail_content)

        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj.load_pipeline_config()
        result = lint_obj.nextflow_config()
        assert len(result["failed"]) == 0
        assert len(result["warned"]) == 0
        assert "Config default value correct: params.dummy" in str(result["passed"])

    def test_default_values_float_fail(self):
        """Test comparing two float values."""
        # Add a float value `dummy=0.0001` to the nextflow.config below `validate_params`
        nf_conf_file = Path(self.new_pipeline) / "nextflow.config"
        with open(nf_conf_file) as f:
            content = f.read()
            fail_content = re.sub(
                r"validate_params\s*=\s*true",
                "params.validate_params = true\ndummy = 0.000000001",
                content,
            )
        with open(nf_conf_file, "w") as f:
            f.write(fail_content)
        # Add a float value `dummy` to the nextflow_schema.json
        nf_schema_file = Path(self.new_pipeline) / "nextflow_schema.json"
        with open(nf_schema_file) as f:
            content = f.read()
            fail_content = re.sub(
                r'"validate_params": {',
                '    "dummy": {"type": "float","default":0.000001},\n"validate_params": {',
                content,
            )
        with open(nf_schema_file, "w") as f:
            f.write(fail_content)

        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj.load_pipeline_config()
        result = lint_obj.nextflow_config()

        assert len(result["failed"]) == 1
        assert len(result["warned"]) == 0
        assert "Config default value incorrect: `params.dummy" in str(result["failed"])
